import numpy as np
import pandas as pd
import pytest

from pydisagg.ihme.validator import (
    validate_columns,
    validate_index,
    validate_interval,
    validate_noindexdiff,
    validate_nonan,
    validate_pat_coverage,
    validate_positive,
    validate_set_uniqueness,
    validate_realnumber,
)


# Test functions
@pytest.fixture
def data():
    np.random.seed(123)
    return pd.DataFrame(
        dict(
            uid=range(10),
            sex_id=[1] * 5 + [2] * 5,
            location_id=[1, 2] * 5,
            year_id=[2010] * 10,
            age_start=[0, 5, 10, 17, 20] * 2,
            age_end=[12, 10, 22, 21, 25] * 2,
            val=5.0,
            val_sd=1.0,
        )
    )


@pytest.fixture
def pattern():
    np.random.seed(123)
    return pd.DataFrame(
        dict(
            sex_id=[1] * 5 + [2] * 5,
            age_start=[0, 5, 10, 15, 20] * 2,
            age_end=[5, 10, 15, 20, 25] * 2,
            age_group_id=list(range(5)) * 2,
            draw_0=np.random.rand(10),
            draw_1=np.random.rand(10),
            draw_2=np.random.rand(10),
            year_id=[2010] * 10,
            location_id=[1, 2] * 5,
        )
    )


@pytest.fixture
def population():
    np.random.seed(123)
    sex_id = pd.DataFrame(dict(sex_id=[1, 2]))
    year_id = pd.DataFrame(dict(year_id=[2010]))
    location_id = pd.DataFrame(dict(location_id=[1, 2]))
    age_group_id = pd.DataFrame(dict(age_group_id=range(5)))

    population = (
        sex_id.merge(location_id, how="cross")
        .merge(age_group_id, how="cross")
        .merge(year_id, how="cross")
    )
    population["population"] = 1000
    return population


# Tests for validate_columns
def test_validate_columns_missing(population):
    with pytest.raises(KeyError):
        validate_columns(
            population.drop(columns=["population"]),
            ["sex_id", "location_id", "age_group_id", "year_id", "population"],
            "population",
        )


def test_validate_columns_no_missing(population):
    # All columns are present; should pass
    validate_columns(
        population,
        ["sex_id", "location_id", "age_group_id", "year_id", "population"],
        "population",
    )


# Tests for validate_index
def test_validate_index_missing(population):
    with pytest.raises(ValueError):
        validate_index(
            pd.concat([population, population]),
            ["sex_id", "location_id", "age_group_id", "year_id"],
            "population",
        )


def test_validate_index_no_duplicates(population):
    # Ensure DataFrame has no duplicate indices; should pass
    validate_index(
        population,
        ["sex_id", "location_id", "age_group_id", "year_id"],
        "population",
    )


# Tests for validate_nonan
def test_validate_nonan(population):
    with pytest.raises(ValueError):
        validate_nonan(population.assign(population=np.nan), "population")


def test_validate_nonan_no_nan(population):
    # No NaN values; should pass
    validate_nonan(population, "population")


# Tests for validate_positive
def test_validate_positive_strict(population):
    with pytest.raises(ValueError):
        validate_positive(
            population.assign(population=0),
            ["population"],
            "population",
            strict=True,
        )


def test_validate_positive_not_strict(population):
    with pytest.raises(ValueError):
        validate_positive(
            population.assign(population=-1),
            ["population"],
            "population",
            strict=False,
        )


def test_validate_positive_no_error(population):
    validate_positive(population, ["population"], "population", strict=True)
    validate_positive(population, ["population"], "population", strict=False)


# Tests for validate_interval
def test_validate_interval_lower_equal_upper(data):
    with pytest.raises(ValueError):
        validate_interval(
            data.assign(age_end=data["age_start"]),
            "age_start",
            "age_end",
            ["uid"],
            "data",
        )


def test_validate_interval_lower_greater_than_upper(data):
    with pytest.raises(ValueError):
        validate_interval(
            data.assign(age_end=0), "age_start", "age_end", ["uid"], "data"
        )


def test_validate_interval_positive(data):
    validate_interval(data, "age_start", "age_end", ["uid"], "data")


# Tests for validate_noindexdiff
@pytest.fixture
def merged_data_pattern(data, pattern):
    return pd.merge(
        pattern,
        data,
        on=["sex_id", "age_start", "age_end", "location_id"],
        how="outer",
    )


def test_validate_noindexdiff_merged_positive(merged_data_pattern, population):
    # Positive test case: no index difference
    validate_noindexdiff(
        population,
        merged_data_pattern.dropna(subset=["sex_id", "location_id"]),
        ["sex_id", "location_id"],
        "merged_data_pattern",
    )


def test_validate_noindexdiff_merged_negative(data, pattern):
    # Negative test case: index difference exists
    data_with_pattern = data.merge(
        pattern,
        on=["sex_id", "age_start", "age_end", "location_id", "year_id"],
        how="left",
    )
    # Introduce an index difference
    data_with_pattern = data_with_pattern.drop(index=0)
    with pytest.raises(ValueError):
        validate_noindexdiff(
            data,
            data_with_pattern,
            data.columns.tolist(),
            "merged_data_pattern",
        )


# Tests for validate_pat_coverage
@pytest.mark.parametrize(
    "bad_data_with_pattern",
    [
        # gap in pattern
        pd.DataFrame(
            dict(
                pat_lwr=[0, 1, 2, 2, 4, 5],
                pat_upr=[1, 2, 5, 3, 5, 7],
            )
        ),
        # overlap in pattern
        pd.DataFrame(
            dict(
                pat_lwr=[0, 0, 2, 2, 3, 5],
                pat_upr=[1, 2, 5, 3, 5, 7],
            )
        ),
        # doesn't cover head
        pd.DataFrame(
            dict(
                pat_lwr=[1, 2, 3, 2, 3, 5],
                pat_upr=[2, 3, 5, 3, 5, 7],
            )
        ),
        # doesn't cover tail
        pd.DataFrame(
            dict(
                pat_lwr=[0, 1, 2, 2, 3, 5],
                pat_upr=[1, 2, 4, 3, 5, 7],
            )
        ),
    ],
)
def test_validate_pat_coverage_failure(bad_data_with_pattern):
    bad_data_with_pattern["group_id"] = [1, 1, 1, 2, 2, 2]
    bad_data_with_pattern["lwr"] = [0, 0, 0, 3, 3, 3]
    bad_data_with_pattern["upr"] = [5, 5, 5, 7, 7, 7]

    with pytest.raises(ValueError):
        validate_pat_coverage(
            bad_data_with_pattern,
            "lwr",
            "upr",
            "pat_lwr",
            "pat_upr",
            ["group_id"],
            "pattern",
        )


# Tests for validate_realnumber
def test_validate_realnumber_positive():
    df = pd.DataFrame({"col1": [1, 2.5, -3.5, 4.2], "col2": [5.1, 6, 7, 8]})
    # Should pass without exceptions
    validate_realnumber(df, ["col1", "col2"], "df")


def test_validate_realnumber_zero():
    df = pd.DataFrame({"col1": [1, 2, 0, 4], "col2": [5, 6, 7, 8]})
    with pytest.raises(
        ValueError, match="df has non-real or zero values in: \\['col1'\\]"
    ):
        validate_realnumber(df, ["col1"], "df")


def test_validate_realnumber_nan():
    df = pd.DataFrame({"col1": [1, 2, 3, np.nan], "col2": [5, 6, 7, 8]})
    with pytest.raises(
        ValueError, match="df has non-real or zero values in: \\['col1'\\]"
    ):
        validate_realnumber(df, ["col1"], "df")


def test_validate_realnumber_non_numeric():
    df = pd.DataFrame({"col1": [1, 2, 3, "a"], "col2": [5, 6, 7, 8]})
    with pytest.raises(
        ValueError, match="df has non-real or zero values in: \\['col1'\\]"
    ):
        validate_realnumber(df, ["col1"], "df")


def test_validate_realnumber_infinite():
    df = pd.DataFrame({"col1": [1, 2, 3, np.inf], "col2": [5, 6, 7, 8]})
    # np.inf is not a finite real number
    with pytest.raises(
        ValueError, match="df has non-real or zero values in: \\['col1'\\]"
    ):
        validate_realnumber(df, ["col1"], "df")


# Tests for validate_set_uniqueness
def test_validate_set_uniqueness_positive():
    df = pd.DataFrame(
        {"col1": [[1, 2, 3], ["a", "b", "c"], [True, False], [1.1, 2.2, 3.3]]}
    )
    # Should pass without exceptions
    validate_set_uniqueness(df, "col1", "df")


def test_validate_set_uniqueness_negative():
    df = pd.DataFrame(
        {"col1": [[1, 2, 2], ["a", "b", "a"], [True, False], [1.1, 2.2, 1.1]]}
    )
    with pytest.raises(
        ValueError,
        match="df has rows in column 'col1' where list elements are not unique.",
    ):
        validate_set_uniqueness(df, "col1", "df")


def test_validate_set_uniqueness_empty_lists():
    df = pd.DataFrame({"col1": [[], [], []]})
    # Should pass; empty lists have no duplicates
    validate_set_uniqueness(df, "col1", "df")


def test_validate_set_uniqueness_single_element_lists():
    df = pd.DataFrame({"col1": [[1], ["a"], [True]]})
    # Should pass; single-element lists can't have duplicates
    validate_set_uniqueness(df, "col1", "df")


def test_validate_set_uniqueness_mixed_types_with_duplicates():
    df = pd.DataFrame({"col1": [[1, "1", 1.0], [True, 1, 1.0], [2, 2, 2]]})
    with pytest.raises(ValueError):
        validate_set_uniqueness(df, "col1", "df")



from pydisagg.models.disagg_model import DisaggModel
from pydisagg.transformations import LogModifiedOdds


class LMOModel(DisaggModel):
    """DisaggModel using the log-modified odds transformation with the exponent m."""

    def __init__(self, m: float) -> None:
        super().__init__(transformation=LogModifiedOdds(m))



import pytest
from numpy.testing import assert_approx_equal

import pydisagg.transformations as tr

transformations = [
    tr.Log(),
    tr.LogModifiedOdds(1),
    tr.LogModifiedOdds(4),
    tr.LogOdds(),
]

x_values = [0.02, 0.01, 0.5, 0.98]


@pytest.mark.parametrize("T", transformations)
@pytest.mark.parametrize("x", x_values)
def test_inverse_consistency(T, x):
    assert_approx_equal(x, T.inverse(T(x)))


@pytest.mark.parametrize("T", transformations)
@pytest.mark.parametrize("x", x_values)
def test_approximate_derivative(T, x, h=0.001):
    """
    Sanity check that the derivatives are are close to correct with a
    finite difference. It would have been a good idea to just do
    everything with Jax from the start.
    """
    assert_approx_equal(T.diff(x), (T(x + h) - T(x - h)) / (2 * h), 2)



import pytest
import pandas as pd
from pydisagg.ihme.splitter import (
    SexSplitter,
    SexDataConfig,
    SexPatternConfig,
    SexPopulationConfig,
)


# Step 1: Setup Fixtures
@pytest.fixture
def sex_data_config():
    return SexDataConfig(
        index=["age_group_id", "year_id", "location_id"],
        val="val",
        val_sd="val_sd",
    )


@pytest.fixture
def sex_pattern_config():
    return SexPatternConfig(by=["age_group_id", "year_id"])


@pytest.fixture
def sex_population_config():
    return SexPopulationConfig(
        index=["age_group_id", "year_id", "location_id"],
        sex="sex_id",
        sex_m=1,
        sex_f=2,
        val="population",
    )


@pytest.fixture
def valid_data():
    return pd.DataFrame(
        {
            "age_group_id": [1, 1, 2, 2],
            "year_id": [2000, 2000, 2001, 2001],
            "location_id": [10, 20, 10, 20],
            "sex_id": [3, 3, 3, 3],
            "val": [100, 200, 150, 250],
            "val_sd": [10, 20, 15, 25],
        }
    )


@pytest.fixture
def valid_pattern():
    return pd.DataFrame(
        {
            "age_group_id": [1, 1, 2, 2],
            "year_id": [2000, 2000, 2001, 2001],
            "pattern_val": [1.5, 2.0, 1.2, 1.8],
            "pattern_val_sd": [0.1, 0.2, 0.15, 0.25],
        }
    )


@pytest.fixture
def invalid_pattern_missing_columns():
    return pd.DataFrame(
        {
            "age_group_id": [1, 1, 2, 2],
            "year_id": [2000, 2000, 2001, 2001],
            "pattern_val": [1.5, 2.0, 1.2, 1.8],
            # Missing pattern_val_sd
        }
    )


@pytest.fixture
def duplicated_index_pattern():
    return pd.DataFrame(
        {
            "age_group_id": [1, 1, 1, 1],
            "year_id": [2000, 2000, 2000, 2000],
            "pattern_val": [1.5, 2.0, 1.2, 1.8],
            "pattern_val_sd": [0.1, 0.2, 0.15, 0.25],
        }
    )


@pytest.fixture
def pattern_with_nan():
    return pd.DataFrame(
        {
            "age_group_id": [1, 1, 2, 2],
            "year_id": [2000, 2000, 2001, 2001],
            "pattern_val": [1.5, None, 1.2, 1.8],
            "pattern_val_sd": [0.1, 0.2, None, 0.25],
        }
    )


@pytest.fixture
def pattern_with_non_positive():
    return pd.DataFrame(
        {
            "age_group_id": [1, 1, 2, 2],
            "year_id": [2000, 2000, 2001, 2001],
            "pattern_val": [-1.5, 0, -1.2, 0],
            "pattern_val_sd": [0.1, 0.2, 0.15, 0.25],
        }
    )


@pytest.fixture
def sex_splitter(sex_data_config, sex_pattern_config, sex_population_config):
    return SexSplitter(
        data=sex_data_config,
        pattern=sex_pattern_config,
        population=sex_population_config,
    )


# Step 2: Write Tests for parse_data
def test_parse_data_missing_columns(sex_splitter, valid_data):
    """Test parse_data raises an error when columns are missing."""
    invalid_data = valid_data.drop(columns=["val"])
    with pytest.raises(KeyError):
        sex_splitter.parse_data(invalid_data)


def test_parse_data_duplicated_index(sex_splitter, valid_data):
    """Test parse_data raises an error on duplicated index."""
    duplicated_data = pd.concat([valid_data, valid_data])
    with pytest.raises(ValueError):
        sex_splitter.parse_data(duplicated_data)


def test_parse_data_with_nan(sex_splitter, valid_data):
    """Test parse_data raises an error when there are NaN values."""
    nan_data = valid_data.copy()
    nan_data.loc[0, "val"] = None
    with pytest.raises(ValueError):
        sex_splitter.parse_data(nan_data)


def test_parse_data_non_positive(sex_splitter, valid_data):
    """Test parse_data raises an error for non-positive values in val or val_sd."""
    non_positive_data = valid_data.copy()
    non_positive_data.loc[0, "val"] = -10
    with pytest.raises(ValueError):
        sex_splitter.parse_data(non_positive_data)


def test_parse_data_valid(sex_splitter, valid_data):
    """Test that parse_data works correctly on valid data."""
    parsed_data = sex_splitter.parse_data(valid_data)
    assert not parsed_data.empty
    assert "val" in parsed_data.columns
    assert "val_sd" in parsed_data.columns


# Step 3: Write Tests for parse_pattern
def test_parse_pattern_missing_columns(
    sex_splitter, valid_data, invalid_pattern_missing_columns
):
    """Test parse_pattern raises an error when required columns are missing."""
    with pytest.raises(ValueError):
        sex_splitter.parse_pattern(
            valid_data, invalid_pattern_missing_columns, model="rate"
        )


def test_parse_pattern_non_positive(
    sex_splitter, valid_data, pattern_with_non_positive
):
    """Test parse_pattern raises an error for non-positive values in pattern."""
    with pytest.raises(ValueError):
        sex_splitter.parse_pattern(
            valid_data, pattern_with_non_positive, model="rate"
        )


# Step 4: Write Tests for parse_population
def test_parse_population_valid(sex_splitter, valid_data):
    """Test parse_population raises an error when population data is missing for some sex_id values."""
    valid_population = pd.DataFrame(
        {
            "age_group_id": [1, 1, 2, 2],
            "year_id": [2000, 2000, 2001, 2001],
            "location_id": [10, 20, 10, 20],
            "sex_id": [1, 2, 1, 2],  # No sex_id = 3 here
            "population": [1000, 1100, 1200, 1300],
        }
    )

    # Ensure that the function raises an error due to unmatched sex_id values
    with pytest.raises(ValueError):
        sex_splitter.parse_population(valid_data, valid_population)


def test_parse_population_missing_columns(sex_splitter, valid_data):
    """Test parse_population raises an error when required columns are missing."""
    invalid_population = pd.DataFrame(
        {
            "age_group_id": [1, 1, 2, 2],
            "year_id": [2000, 2000, 2001, 2001],
            "location_id": [10, 20, 10, 20],
            "population": [1000, 1100, 1200, 1300],
            # Missing 'sex_id' column
        }
    )
    with pytest.raises(KeyError):
        sex_splitter.parse_population(valid_data, invalid_population)


def test_parse_population_with_nan(sex_splitter, valid_data):
    """Test parse_population raises an error when there are NaN values."""
    invalid_population = pd.DataFrame(
        {
            "age_group_id": [1, 1, 2, 2],
            "year_id": [2000, 2000, 2001, 2001],
            "location_id": [10, 20, 10, 20],
            "sex_id": [1, 2, 1, 2],
            "population": [
                1000,
                None,
                1200,
                1300,
            ],  # NaN value in the population
        }
    )
    with pytest.raises(ValueError):
        sex_splitter.parse_population(valid_data, invalid_population)


def test_parse_population_duplicated_index(sex_splitter, valid_data):
    """Test parse_population raises an error when population index is duplicated."""
    duplicated_population = pd.DataFrame(
        {
            "age_group_id": [1, 1, 1, 1],
            "year_id": [2000, 2000, 2000, 2000],
            "location_id": [10, 20, 10, 20],
            "sex_id": [1, 2, 1, 2],
            "population": [1000, 1100, 1200, 1300],
        }
    )
    with pytest.raises(ValueError):
        sex_splitter.parse_population(valid_data, duplicated_population)



import numpy as np
import pytest
import pandas as pd

from pydisagg import models
from pydisagg.disaggregate import split_datapoint, split_dataframe

model_list = [
    models.RateMultiplicativeModel(),
    models.LogOddsModel(),
    models.LMOModel(5),
]


# 1. Test split_datapoint with invalid output_type
def test_split_datapoint_invalid_output_type():
    populations = np.array([2, 5])
    rate_pattern = np.array([0.2, 0.4])
    with pytest.raises(ValueError):
        split_datapoint(4.8, populations, rate_pattern, output_type="invalid")


# 2. Test split_datapoint without observed_total_se
@pytest.mark.parametrize("model", model_list)
def test_split_datapoint_no_se(model):
    populations = np.array([2, 5])
    rate_pattern = np.array([0.2, 0.4])
    result = split_datapoint(4.8, populations, rate_pattern, model=model)
    assert result is not None


# 3. Test split_datapoint with pattern_covariance
@pytest.mark.parametrize("model", model_list)
def test_split_datapoint_with_covariance(model):
    populations = np.array([2, 5])
    rate_pattern = np.array([0.2, 0.4])
    covariance = np.array([[0.1, 0.02], [0.02, 0.1]])
    result, SE = split_datapoint(
        4.8,
        populations,
        rate_pattern,
        1,
        model=model,
        pattern_covariance=covariance,
    )
    assert result is not None
    assert SE is not None


# 4. Test split_dataframe with mismatched index
def test_split_dataframe_mismatched_index():
    groups_to_split = ["A", "B"]
    obs_df = pd.DataFrame(
        {
            "demographic_id": [1, 2],
            "pattern_id": [1, 1],
            "obs": [100, 150],
            "A": [1, 0],
            "B": [0, 1],
        }
    )

    pop_df = pd.DataFrame({"A": [50, 75], "B": [50, 75]}, index=[3, 4])

    rate_df = pd.DataFrame({"A": [0.2, 0.3], "B": [0.4, 0.5]}, index=[1, 1])

    with pytest.raises(KeyError):
        split_dataframe(groups_to_split, obs_df, pop_df, rate_df)


# 5. Test split_dataframe with empty dataframes
def test_split_dataframe_empty():
    groups_to_split = ["A", "B"]
    obs_df = pd.DataFrame(
        columns=["demographic_id", "pattern_id", "obs", "A", "B"]
    )
    pop_df = pd.DataFrame(columns=["A", "B"])
    rate_df = pd.DataFrame(columns=["A", "B"])

    result = split_dataframe(groups_to_split, obs_df, pop_df, rate_df)
    assert result.empty


def test_split_datapoint_output_type_rate():
    populations = np.array([2, 5])
    rate_pattern = np.array([0.2, 0.4])
    result = split_datapoint(4.8, populations, rate_pattern, output_type="rate")
    assert result is not None
    assert len(result) == 2


def test_split_datapoint_with_observed_total_se():
    populations = np.array([2, 5])
    rate_pattern = np.array([0.2, 0.4])
    observed_total_se = 1.0
    result, se = split_datapoint(
        4.8, populations, rate_pattern, observed_total_se=observed_total_se
    )
    assert result is not None
    assert se is not None


def test_split_datapoint_negative_observed_total():
    populations = np.array([2, 5])
    rate_pattern = np.array([0.2, 0.4])
    with pytest.raises(ValueError):
        split_datapoint(-4.8, populations, rate_pattern)


def test_split_datapoint_zero_population():
    populations = np.array([0, 5])
    rate_pattern = np.array([0.2, 0.4])
    result = split_datapoint(4.8, populations, rate_pattern)
    assert result is not None


def test_split_datapoint_negative_population():
    populations = np.array([-2, 5])
    rate_pattern = np.array([0.2, 0.4])
    with pytest.raises(ValueError):
        split_datapoint(4.8, populations, rate_pattern)


def test_split_datapoint_zero_rate_pattern():
    populations = np.array([2, 5])
    rate_pattern = np.array([0.0, 0.4])
    result = split_datapoint(4.8, populations, rate_pattern)
    assert result is not None


def test_split_datapoint_empty_populations():
    populations = np.array([])
    rate_pattern = np.array([0.2, 0.4])
    with pytest.raises(ValueError):
        split_datapoint(4.8, populations, rate_pattern)


def test_split_datapoint_empty_rate_pattern():
    populations = np.array([2, 5])
    rate_pattern = np.array([])
    with pytest.raises(ValueError):
        split_datapoint(4.8, populations, rate_pattern)


def test_split_datapoint_mismatched_lengths():
    populations = np.array([2, 5, 7])
    rate_pattern = np.array([0.2, 0.4])
    with pytest.raises(ValueError):
        split_datapoint(4.8, populations, rate_pattern)


def test_split_datapoint_non_numeric_populations():
    populations = np.array([2, "a"])
    rate_pattern = np.array([0.2, 0.4])
    with pytest.raises(ValueError):
        split_datapoint(4.8, populations, rate_pattern)


def test_split_datapoint_non_numeric_rate_pattern():
    populations = np.array([2, 5])
    rate_pattern = np.array([0.2, "b"])
    with pytest.raises(ValueError):
        split_datapoint(4.8, populations, rate_pattern)



"""
Module containing classes of Transformations
"""

from abc import ABC, abstractmethod

import numpy as np
from scipy.optimize import root_scalar

from pydisagg.typing import float_or_array


class Transformation(ABC):
    """Transformation that contains function, first order and inverse
    information
    """

    @abstractmethod
    def __call__(self, x: float_or_array) -> float_or_array:
        """Calls transformation function"""

    @abstractmethod
    def inverse(self, z: float_or_array) -> float_or_array:
        """Calls inverse of transformation function"""

    @abstractmethod
    def diff(self, x: float_or_array) -> float_or_array:
        """Calls derivative of transformation"""


class Log(Transformation):
    """Logarithmic transformation of parameter
    Since we fit models additively in beta the log makes it multiplicative
    """

    def __call__(self, x: float_or_array) -> float_or_array:
        return np.log(x)

    def inverse(self, z: float_or_array) -> float_or_array:
        return np.exp(z)

    def diff(self, x: float_or_array) -> float_or_array:
        return 1 / x


class LogModifiedOdds(Transformation):
    """Log Modified odds transformation T(x)=log(x/(1-x**a))"""

    def __init__(self, a: float) -> None:
        self.a = a

    def __call__(self, x: float_or_array) -> float_or_array:
        """Calls transformation function"""
        return np.log(x / (1 - (x**self.a)))

    def _inverse_single(self, z: float) -> float:
        def root_func(x):
            return np.exp(z) * (1 - x**self.a) - x

        return root_scalar(root_func, bracket=[0, 1], method="toms748").root

    def inverse(self, z: float_or_array) -> float_or_array:
        return np.vectorize(self._inverse_single)(z)

    def diff(self, x: float_or_array) -> float_or_array:
        numerator = (self.a - 1) * (x**self.a) + 1
        denominator = ((x**self.a) - 1) * x
        return -1 * numerator / denominator


class LogOdds(Transformation):
    """Log-odds transformation T(x)=log(x/(1-x))"""

    def __call__(self, x: float_or_array) -> float_or_array:
        """Calls transformation function"""
        return np.log(x / (1 - x))

    def inverse(self, z: float_or_array) -> float_or_array:
        expz = np.exp(z)
        return expz / (1 + expz)

    def diff(self, x: float_or_array) -> float_or_array:
        return 1 / (x - x**2)



"""Module containing high level api for splitting"""

from typing import Literal

import numpy as np
import pandas as pd

from pydisagg.models import DisaggModel, LogOddsModel
from pydisagg.typing import DataFrame, NDArray


def split_datapoint(
    observed_total: float,
    bucket_populations: NDArray,
    rate_pattern: NDArray,
    observed_total_se: float | None = None,
    model: DisaggModel = LogOddsModel(),
    output_type: Literal["count", "rate"] = "count",
    normalize_pop_for_average_type_obs: bool = False,
    pattern_covariance: NDArray | None = None,
) -> tuple | NDArray:
    """Disaggregate a datapoint using the model given as input.
    Defaults to assuming multiplicativity in the odds ratio

    * If output_type=='total', then this outputs estimates for the observed
      amount in each group such that the sum of the point estimates equals the
      original total
    * If output_type=='rate', then this estimates rates for each group
      (and doesn't multiply the rates out by the population)


    Parameters
    ----------
    observed_total
        aggregated observed_total across all buckets, value to be split
    bucket_populations
        population size in each bucket
    rate_pattern
        Rate Pattern to use, should be an estimate of the rates in each bucket
            that we want to rescale
    observed_total_se
        standard error of observed_total, by default None
    output_type
        One of 'total' or 'rate'
        Type of splitting to perform, whether to disaggregate and return the
        estimated total in each group, or estimate the rate per population unit.
    model
        DisaggModel to use, by default LMOModel(1)
    normalize_pop_for_average_type_obs
        Whether or not to normalize populations to sum to 1, this is appropriate
        when the output_type is rate and when the aggregated observation is an
        average--whether an aggregated rate or a mean of a continuous measure
        over different groups
    pattern_covariance
        2d Numpy array with covariance matrix of pattern.

    Returns
    -------
    tuple | NDArray
        If standard errors are available, this will return the tuple
            (
                estimate_in_each_bucket,
                se_of_estimate_bucket,
            )
        Otherwise, if standard errors are not available,
        this will return a numpy array of the disaggregated estimates

    Notes
    -----
    If no observed_total_se is given, returns point estimates
    If observed_total_se is given, then returns a tuple
        (point_estimate,standard_error)

    """
    if not np.issubdtype(bucket_populations.dtype, np.number):
        raise ValueError("All elements in bucket_populations must be numeric")
    if not np.issubdtype(rate_pattern.dtype, np.number):
        raise ValueError("All elements in rate_pattern must be numeric")

    if output_type not in ["count", "rate"]:
        raise ValueError("output_type must be one of either 'count' or 'rate'")

    if normalize_pop_for_average_type_obs is True:
        processed_bucket_populations = bucket_populations / np.sum(
            bucket_populations
        )
    else:
        processed_bucket_populations = bucket_populations.copy()

    if output_type == "count":
        point_estimates = model.split_to_counts(
            observed_total, rate_pattern, processed_bucket_populations
        )
        if (observed_total_se is not None) and (pattern_covariance is None):
            fitted_beta = model.fit_beta(
                observed_total, rate_pattern, processed_bucket_populations
            )
            standard_errors = model.count_split_standard_errors(
                fitted_beta,
                rate_pattern,
                processed_bucket_populations,
                observed_total_se,
            )
            return point_estimates, standard_errors

        if (observed_total_se is not None) and (pattern_covariance is not None):
            fitted_beta = model.fit_beta(
                observed_total, rate_pattern, processed_bucket_populations
            )

            cov_mat = model.count_split_covariance_uncertainty(
                fitted_beta,
                rate_pattern,
                processed_bucket_populations,
                observed_total_se,
                pattern_covariance,
            )
            standard_errors = np.sqrt(np.diag(cov_mat))

            return point_estimates, standard_errors

        return point_estimates

    if output_type == "rate":
        point_estimates = model.split_to_rates(
            observed_total,
            rate_pattern,
            processed_bucket_populations,
            reduce_output=True,
        )
        if (observed_total_se is not None) and (pattern_covariance is None):
            fitted_beta = model.fit_beta(
                observed_total, rate_pattern, processed_bucket_populations
            )
            standard_errors = model.rate_standard_errors(
                fitted_beta,
                rate_pattern,
                processed_bucket_populations,
                observed_total_se,
            )
            return point_estimates, standard_errors

        if (observed_total_se is not None) and (pattern_covariance is not None):
            fitted_beta = model.fit_beta(
                observed_total, rate_pattern, processed_bucket_populations
            )

            cov_mat = model.rate_split_covariance_uncertainty(
                fitted_beta,
                rate_pattern,
                processed_bucket_populations,
                observed_total_se,
                pattern_covariance,
            )
            standard_errors = np.sqrt(np.diag(cov_mat))

            return point_estimates, standard_errors

        return point_estimates


def split_dataframe(
    groups_to_split_into: list,
    observation_group_membership_df: DataFrame,
    population_sizes: DataFrame,
    rate_patterns: DataFrame,
    use_se: bool = False,
    model: DisaggModel = LogOddsModel(),
    output_type: Literal["count", "rate"] = "count",
    demographic_id_columns: list | None = None,
    normalize_pop_for_average_type_obs: bool = False,
) -> DataFrame:
    """Disaggregate datapoints and pivots observations into estimates for each
    group per demographic id

    * If output_type=='total', then this outputs estimates for the observed
      amount in each group such that the sum of the point estimates equals the
      original total
    * If output_type=='rate', then this estimates rates for each group
      (and doesn't multiply the rates out by the population)

    Parameters
    ----------
    groups_to_split_into
        list of groups to disaggregate observations into
    observation_group_membership_df
        Dataframe with columns demographic_id, pattern_id, obs,
        and columns for each of the groups_to_split_into
        with dummy variables that represent whether or not
        each group is included in the observations for that row.
        This also optionally contains a obs_se column which will be used if
        use_se is True. demographic_id represents the population that the
        observation comes from pattern_id gives the baseline that should be used
        for splitting
    population_sizes
        Dataframe with demographic_id as the index containing the
        size of each group within each population (given the demographic_id)
        INDEX FOR THIS DATAFRAME MUST BE DEMOGRAPHIC ID(PANDAS MULTIINDEX OK)
    rate_patterns
        dataframe with pattern_id as the index, and columns
        for each of the groups_to_split where the entries represent the rate
        pattern in the given group to use for pydisagg.
    use_se
        whether or not to report standard errors along with estimates
        if set to True, then observation_group_membership_df must have an obs_se
        column, by default False
    model
        DisaggModel to use for splitting, by default LogOddsModel()
    output_type
        One of 'total' or 'rate'
        Type of splitting to perform, whether to disaggregate and return the
        estimated total in each group, or estimate the rate per population unit.
    demographic_id_columns
        Columns to use as demographic_id
        Defaults to None. If None is given, then we assume
        that there is a already a demographic id column that matches the index
        in population_sizes. Otherwise, we create a new demographic_id column,
        zipping the columns chosen into tuples
    normalize_pop_for_average_type_obs
        Whether or not to normalize populations to sum to 1, this is appropriate
        when the output_type is rate and when the aggregated observation is an
        average--whether an aggregated rate or a mean of a continuous measure
        over different groups

    Returns
    -------
    DataFrame
        Dataframe where each row corresponds to one of obs, with one or
            two columns for each of the groups_to_split_into, giving the estimate
        If use_se==True, then has a nested column indexing, where both the
            point estimate and standard error for the estimate for each group is given.

    """
    if (normalize_pop_for_average_type_obs is True) and (
        output_type == "count"
    ):
        raise Warning(
            "Normalizing populations may not be appropriate here, as we are working with a total"
        )

    splitting_df = observation_group_membership_df.copy()
    if demographic_id_columns is not None:
        splitting_df["demographic_id"] = list(
            zip(*[splitting_df[id_col] for id_col in demographic_id_columns])
        )

    if use_se is False:

        def split_row(x):
            return split_datapoint(
                x["obs"],
                population_sizes.loc[x.name] * x[groups_to_split_into],
                rate_patterns.loc[x["pattern_id"]],
                model=model,
                output_type=output_type,
                normalize_pop_for_average_type_obs=normalize_pop_for_average_type_obs,
            )

        result = (
            splitting_df.set_index("demographic_id")
            .apply(split_row, axis=1)
            .reset_index()
        )
    else:

        def split_row(x):
            raw_split_result = split_datapoint(
                x["obs"],
                population_sizes.loc[x.name] * x[groups_to_split_into],
                rate_patterns.loc[x["pattern_id"]],
                model=model,
                observed_total_se=x["obs_se"],
                output_type=output_type,
                normalize_pop_for_average_type_obs=normalize_pop_for_average_type_obs,
            )
            return pd.Series(
                [
                    (estimate, se)
                    for estimate, se in zip(
                        raw_split_result[0], raw_split_result[1]
                    )
                ],
                index=groups_to_split_into,
            )

        result_raw = splitting_df.set_index("demographic_id").apply(
            split_row, axis=1
        )
        point_estimates = result_raw.applymap(lambda x: x[0])
        standard_errors = result_raw.applymap(lambda x: x[1])
        result = pd.concat(
            [point_estimates, standard_errors], keys=["estimate", "se"], axis=1
        ).reset_index()  # .groupby(level=0).sum()

    if demographic_id_columns is not None:
        demographic_id_df = pd.DataFrame(
            dict(zip(demographic_id_columns, zip(*result["demographic_id"]))),
            index=result.index,
        )
        result = pd.concat([demographic_id_df, result], axis=1).drop(
            "demographic_id", axis=1
        )

    return result



import numpy as np
from numpy.typing import NDArray

from pydisagg.models.disagg_model import DisaggModel
from pydisagg.transformations import Log


class RateMultiplicativeModel(DisaggModel):
    """Produces a DisaggModel using the log(rate) transformation with the
    exponent m. This assumes that log(rate)=log(rate_pattern)+beta
    resulting in the current multiplicative model after exponentiating
    Take exp(beta) to recover the multiplier in the model.

    """

    def __init__(self) -> None:
        super().__init__(transformation=Log())

    def fit_beta(
        self,
        observed_total: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
        lower_guess: float = -50,
        upper_guess: float = 50,
        verbose: int = 0,
    ) -> None:
        """
        Custom fit_beta for this model, as we can do it without rootfinding.
        """
        beta_val = np.log(
            observed_total / np.sum(bucket_populations * rate_pattern)
        )
        return beta_val



from typing import Any

import numpy as np
import pandas as pd
from pandas import DataFrame
from pydantic import BaseModel
from typing import Literal
from pydisagg.disaggregate import split_datapoint
from pydisagg.ihme.schema import Schema
from pydisagg.ihme.validator import (
    validate_columns,
    validate_index,
    validate_noindexdiff,
    validate_nonan,
    validate_positive,
    validate_realnumber,
)
from pydisagg.models import RateMultiplicativeModel


class SexPatternConfig(Schema):
    by: list[str]
    draws: list[str] = []
    val: str = "ratio_f_to_m"
    val_sd: str = "ratio_f_to_m_se"
    prefix: str = "sex_pat_"

    @property
    def index(self) -> list[str]:
        return self.by

    @property
    def columns(self) -> list[str]:
        return self.index + [
            self.val,
            self.val_sd,
        ]

    @property
    def val_fields(self) -> list[str]:
        return [
            "val",
            "val_sd",
        ]


class SexPopulationConfig(Schema):
    index: list[str]
    sex: str
    sex_m: str | int
    sex_f: str | int
    val: str
    prefix: str = "sex_pop_"

    @property
    def columns(self) -> list[str]:
        return self.index + [self.sex, self.val]

    @property
    def val_fields(self) -> list[str]:
        return ["val"]


class SexDataConfig(Schema):
    index: list[str]
    val: str
    val_sd: str

    @property
    def columns(self) -> list[str]:
        return list(set(self.index + [self.val, self.val_sd]))


class SexSplitter(BaseModel):
    data: SexDataConfig
    pattern: SexPatternConfig
    population: SexPopulationConfig

    def model_post_init(self, __context: Any) -> None:
        """Extra validation of all the index."""
        if not set(self.pattern.by).issubset(self.data.index):
            raise ValueError("pattern.by must be a subset of data.index")
        if not set(self.population.index).issubset(
            self.data.index + self.pattern.index
        ):
            raise ValueError(
                "population.index must be a subset of data.index + pattern.index"
            )

    def _merge_with_pattern(
        self, data: DataFrame, pattern: DataFrame
    ) -> DataFrame:
        data_with_pattern = data.merge(
            pattern, on=self.pattern.by, how="left"
        ).dropna()
        return data_with_pattern

    def get_population_by_sex(self, population, sex_value):
        return population[population[self.population.sex] == sex_value][
            self.population.index + [self.population.val]
        ].copy()

    def parse_data(self, data: DataFrame) -> DataFrame:
        name = "While parsing data"

        # Validate core columns first
        try:
            validate_columns(data, self.data.columns, name)
        except KeyError as e:
            raise KeyError(
                f"{name}: Missing columns in the input data. Details:\n{e}"
            )

        if self.population.sex not in data.columns:
            raise KeyError(
                f"{name}: Missing column '{self.population.sex}' in the input data."
            )

        try:
            validate_index(data, self.data.index, name)
        except ValueError as e:
            raise ValueError(f"{name}: Duplicated index found. Details:\n{e}")

        try:
            validate_nonan(data, name)
        except ValueError as e:
            raise ValueError(f"{name}: NaN values found. Details:\n{e}")

        try:
            validate_positive(data, [self.data.val, self.data.val_sd], name)
        except ValueError as e:
            raise ValueError(
                f"{name}: Non-positive values found in 'val' or 'val_sd'. Details:\n{e}"
            )

        return data

    def parse_pattern(
        self, data: DataFrame, pattern: DataFrame, model: str
    ) -> DataFrame:
        name = "While parsing pattern"

        try:
            if not all(
                col in pattern.columns
                for col in [self.pattern.val, self.pattern.val_sd]
            ):
                if not self.pattern.draws:
                    raise ValueError(
                        f"{name}: Must provide draws for pattern if pattern.val and "
                        "pattern.val_sd are not available."
                    )
                validate_columns(pattern, self.pattern.draws, name)
                pattern[self.pattern.val] = pattern[self.pattern.draws].mean(
                    axis=1
                )
                pattern[self.pattern.val_sd] = pattern[self.pattern.draws].std(
                    axis=1
                )

            validate_columns(pattern, self.pattern.columns, name)
        except KeyError as e:
            raise KeyError(
                f"{name}: Missing columns in the pattern. Details:\n{e}"
            )

        pattern = pattern[self.pattern.columns].copy()

        try:
            validate_index(pattern, self.pattern.index, name)
        except ValueError as e:
            raise ValueError(
                f"{name}: Duplicated index found in the pattern. Details:\n{e}"
            )

        try:
            validate_nonan(pattern, name)
        except ValueError as e:
            raise ValueError(
                f"{name}: NaN values found in the pattern. Details:\n{e}"
            )

        if model == "rate":
            try:
                validate_positive(
                    pattern, [self.pattern.val, self.pattern.val_sd], name
                )
            except ValueError as e:
                raise ValueError(
                    f"{name}: Non-positive values found in 'val' or 'val_sd'. Details:\n{e}"
                )
        elif model == "logodds":
            try:
                validate_realnumber(pattern, [self.pattern.val_sd], name)
            except ValueError as e:
                raise ValueError(
                    f"{name}: Invalid real number values found. Details:\n{e}"
                )

        pattern_copy = pattern.copy()
        rename_map = self.pattern.apply_prefix()
        pattern_copy.rename(columns=rename_map, inplace=True)

        data_with_pattern = self._merge_with_pattern(data, pattern_copy)

        # Validate index differences after merging
        validate_noindexdiff(data, data_with_pattern, self.data.index, name)

        return data_with_pattern

    def parse_population(
        self, data: DataFrame, population: DataFrame
    ) -> DataFrame:
        name = "While parsing population"

        # Step 1: Validate population columns
        try:
            validate_columns(population, self.population.columns, name)
        except KeyError as e:
            raise KeyError(
                f"{name}: Missing columns in the population data. Details:\n{e}"
            )

        # Step 2: Get male and female populations and rename columns
        male_population = self.get_population_by_sex(
            population, self.population.sex_m
        )
        female_population = self.get_population_by_sex(
            population, self.population.sex_f
        )

        male_population.rename(
            columns={self.population.val: "m_pop"}, inplace=True
        )
        female_population.rename(
            columns={self.population.val: "f_pop"}, inplace=True
        )

        # Step 3: Merge population data with main data
        data_with_population = self._merge_with_population(
            data, male_population, "m_pop"
        )
        data_with_population = self._merge_with_population(
            data_with_population, female_population, "f_pop"
        )

        # Step 4: Validate the merged data columns
        try:
            validate_columns(data_with_population, ["m_pop", "f_pop"], name)
        except KeyError as e:
            raise KeyError(
                f"{name}: Missing population columns after merging. Details:\n{e}"
            )

        # Step 5: Validate for NaN values in the merged columns using validate_nonan
        try:
            validate_nonan(data_with_population, name)
        except ValueError as e:
            raise ValueError(
                f"{name}: NaN values found in the population data. Details:\n{e}"
            )

        # Step 6: Validate index differences
        try:
            validate_noindexdiff(
                data, data_with_population, self.data.index, name
            )
        except ValueError as e:
            raise ValueError(
                f"{name}: Index differences found between data and population. Details:\n{e}"
            )

        # Ensure the columns are in the correct numeric type (e.g., float64)
        data_with_population["m_pop"] = data_with_population["m_pop"].astype(
            "float64"
        )
        data_with_population["f_pop"] = data_with_population["f_pop"].astype(
            "float64"
        )

        return data_with_population

    def _merge_with_population(
        self, data: DataFrame, population: DataFrame, pop_col: str
    ) -> DataFrame:
        keep_cols = self.population.index + [pop_col]
        population_temp = population[keep_cols]
        data_with_population = data.merge(
            population_temp,
            on=self.population.index,
            how="left",
        )

        # Ensure the merged population columns are standard numeric types
        if pop_col in data_with_population.columns:
            data_with_population[pop_col] = data_with_population[
                pop_col
            ].astype("float64")

        return data_with_population

    def split(
        self,
        data: DataFrame,
        pattern: DataFrame,
        population: DataFrame,
        model: Literal["rate", "logodds"] = "rate",
        output_type: Literal["rate", "count"] = "rate",
    ) -> DataFrame:
        """
        Split the input data based on a specified pattern and population model.

        This function splits the input data into male and female groups based on
        the population and pattern data provided. It can operate using either a
        "rate" or "logodds" model for splitting. If the `sex_id` in a row is
        already for a specific sex (male or female), the row's values are
        inherited and not split. Otherwise, the function splits the row's data
        and calculates corresponding standard errors.

        Parameters
        ----------
        data : pd.DataFrame
            The input data to be split. This should include information on the
            observed total value (`self.data.val`) and the associated standard
            error (`self.data.val_sd`), as well as sex identifiers.
        pattern : pd.DataFrame
            The pattern data used to inform the split. This includes pattern
            values (`self.pattern.val`) and associated standard errors
            (`self.pattern.val_sd`) which will be applied during the split.
        population : pd.DataFrame
            Population data that provides the population sizes for male and
            female groups. This is required for proportional splitting of the
            data.
        model : {"rate", "logodds"}, optional
            The model to be used for splitting the data, by default "rate".
            - "rate": Splits based on a multiplicative rate model.
            - "logodds": Splits based on a log-odds model.
        output_type : {"rate", "count"}, optional
            The type of output expected, by default "rate".
            - "rate": The output will be normalized by population.
            - "count": The output will not be normalized by population.

        Returns
        -------
        pd.DataFrame
            A DataFrame with the following columns:
            - `sex_split_result`: The split value for male and female groups.
            - `sex_split_result_se`: The standard error associated with the
            split values.
            - `sex_split`: A flag (0 or 1) indicating whether the row was split.
        """

        # Ensure no prefixes in the pattern config
        if self.pattern.prefix_status == "prefixed":
            self.pattern.remove_prefix()
        if self.population.prefix_status == "prefixed":
            self.population.remove_prefix()

        # Parsing input data, pattern, and population
        data = self.parse_data(data)
        data = self.parse_pattern(data, pattern, model)
        data = self.parse_population(data, population)

        # Determine whether to normalize by population for the output type
        pop_normalize = output_type == "rate"

        # Step 1: Handle rows where `sex_id` is already `sex_m` or `sex_f`
        mask_sex_m_or_f = data[self.population.sex].isin(
            [self.population.sex_m, self.population.sex_f]
        )

        # Create a copy for the final DataFrame where rows are not split
        final_df = data[mask_sex_m_or_f].copy()

        # Set the `sex_split_result`, `sex_split_result_se`, and `sex_split` columns for non-split rows
        final_df["sex_split_result"] = final_df[self.data.val]
        final_df["sex_split_result_se"] = final_df[self.data.val_sd]
        final_df["sex_split"] = 0  # Mark as not split

        # Step 2: Handle rows that need to be split (where `sex_id` is not `sex_m` or `sex_f`)
        split_data = data[~mask_sex_m_or_f].copy()

        # Calculate input patterns and splitting logic
        if model == "rate":
            input_patterns = np.vstack(
                [np.ones(len(split_data)), split_data[self.pattern.val].values]
            ).T
            splitting_model = RateMultiplicativeModel()
        elif model == "logodds":
            raise ValueError(
                "logodds has been removed from version 0.6.2. For sex splitting with bounded quantities, use the cat_splitter function and pass sex specific patterns directly instead of a ratio or logit difference."
            )

        # Perform the split for all rows at once using vectorized operations
        split_results, SEs = zip(
            *split_data.apply(
                lambda row: split_datapoint(
                    observed_total=row[self.data.val],
                    bucket_populations=np.array([row["m_pop"], row["f_pop"]]),
                    rate_pattern=input_patterns[
                        split_data.index.get_loc(row.name)
                    ],
                    model=splitting_model,
                    output_type=output_type,
                    normalize_pop_for_average_type_obs=pop_normalize,
                    observed_total_se=row[self.data.val_sd],
                    pattern_covariance=np.diag(
                        [0, row[self.pattern.val_sd] ** 2]
                    ),
                ),
                axis=1,
            )
        )

        # Split results into male and female values
        split_results = np.array(split_results)
        SEs = np.array(SEs)

        # Step 3: Create male and female dataframes for the split data
        male_split_data = split_data.copy()
        male_split_data[self.population.sex] = self.population.sex_m
        male_split_data["sex_split_result"] = split_results[:, 0]
        male_split_data["sex_split_result_se"] = SEs[:, 0]
        male_split_data["sex_split"] = 1

        female_split_data = split_data.copy()
        female_split_data[self.population.sex] = self.population.sex_f
        female_split_data["sex_split_result"] = split_results[:, 1]
        female_split_data["sex_split_result_se"] = SEs[:, 1]
        female_split_data["sex_split"] = 1

        # Step 4: Combine the non-split rows and the split male/female rows
        final_split_df = pd.concat(
            [final_df, male_split_data, female_split_data], ignore_index=True
        )

        # Reindex columns
        final_split_df = final_split_df.reindex(
            columns=self.data.index
            + [
                col
                for col in final_split_df.columns
                if col not in self.data.index
            ]
        )

        # Clean up any prefixes added earlier
        self.pattern.remove_prefix()
        self.population.remove_prefix()

        return final_split_df



from enum import StrEnum

from pydantic import BaseModel


class PrefixStatus(StrEnum):
    PREFIXED = "prefixed"
    NOT_PREFIXED = "not_prefixed"


class Schema(BaseModel):
    prefix: str = ""
    prefix_status: PrefixStatus = PrefixStatus.NOT_PREFIXED

    @property
    def columns(self) -> list[str]:
        return []

    @property
    def val_fields(self) -> list[str]:
        return []

    def apply_prefix(self) -> dict[str, str]:
        if self.prefix_status == PrefixStatus.PREFIXED:
            raise ValueError("prefix already applied")
        rename_map = {}
        for field in self.val_fields:
            new_field_val = self.prefix + (field_val := getattr(self, field))
            rename_map[field_val] = new_field_val
            setattr(self, field, new_field_val)
        self.prefix_status = PrefixStatus.PREFIXED
        return rename_map

    def remove_prefix(self) -> None:
        for field in self.val_fields:
            setattr(self, field, getattr(self, field).removeprefix(self.prefix))
        self.prefix_status = PrefixStatus.NOT_PREFIXED



"""Module with main DisaggModel class and algorithmic implementation"""

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import root_scalar

from pydisagg.transformations import Transformation


class DisaggModel:
    """Model for solving splitting/disaggregation problems

    transformation is a class attribute that defines the generalized
    proportionality assumption that we want to make
    It should be a Transformation object which is callable, has an
    inverse function, and has a derivative.

    Notes
    -----
    A possibly confusing code and notation choice is that we enforce
    multiplicativity in some space, but fit an additive parameter beta
    throughout

    We do this because it works better with the delta method calculations, and
    we assume that we have a log in the transormations so additive factors
    become multiplicative

    """

    def __init__(self, transformation: Transformation):
        """Initializes a dissaggregation model

        Parameters
        ----------
        transformation
            Transformation to apply to rate pattern values

        """
        self.T = transformation

    def predict_rate(self, beta: float, rate_pattern: NDArray) -> NDArray:
        """Predicts the rate in each bucket

        Parameters
        ----------
        beta : float
            beta paramter, log of rescaling parameter

        rate_pattern : NDArray
            rate pattern

        Returns
        -------
        NDArray
            estimated occurrences in each bucket

        """
        return self.T.inverse(beta + self.T(rate_pattern))

    def rate_diff_beta(self, beta: float, rate_pattern: NDArray) -> NDArray:
        """Computes the derivative of the predicted rates with respect to beta

        Parameters
        ----------
        beta : float
            beta paramter, log of rescaling parameter
        rate_pattern : NDArray
            rate pattern
        bucket_populations :NDArray
            Populations in each bucket
        Returns
        -------
        NDArray
            derivative of predicted rates in each bucket with respect to beta
        """
        return 1 / self.T.diff(self.T.inverse(beta + self.T(rate_pattern)))

    def predict_count(
        self,
        beta: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
    ) -> NDArray:
        """Generate a predicted count in each bucket assuming multiplicativity
        in the T-transformed space with the additive parameter beta

        Parameters
        ----------
        beta : float
            beta paramter, log of rescaling parameter
        rate_pattern : NDArray
            rate pattern
        bucket_populations :NDArray
            Populations in each bucket
        Returns
        -------
        NDArray
            Predicted rates in each bucket
        """
        return self.predict_rate(beta, rate_pattern) * bucket_populations

    def count_diff_beta(
        self,
        beta: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
    ) -> NDArray:
        """Computes the derivative of the predicted count in each bucket with
        respect to beta

        Parameters
        ----------
        beta
            beta paramter, log of rescaling parameter
        rate_pattern
            rate pattern
        bucket_populations
            Populations in each bucket

        Returns
        -------
        NDArray
            Derivative of predicted counts in each bucket with respect to beta

        """
        return self.rate_diff_beta(beta, rate_pattern) * bucket_populations

    def fit_beta(
        self,
        observed_total: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
        lower_guess: float = -50.0,
        upper_guess: float = 50.0,
        verbose: int = 0,
    ) -> float:
        """Fits the parameter beta to the data

        Parameters
        ----------
        observed_total
            aggregated observation
        rate_pattern
            rate pattern to assume (generalized) proportionality to
        bucket_populations
            populations in each bucket
        lower_guess
            Lower bound for rootfinding (we use bracketing), by default -50
        upper_guess
            Upper bound for rootfinding, by default 50
        verbose
            how much to print, 1 prints the root value,
            2 prints the entire rootfinding output, by default 0

        Returns
        -------
        float
            computed value for beta

        """

        def beta_misfit(beta):
            return (
                self.H_func(
                    beta,
                    rate_pattern=rate_pattern,
                    bucket_populations=bucket_populations,
                )
                - observed_total
            )

        beta_results = root_scalar(
            beta_misfit, bracket=[lower_guess, upper_guess], method="toms748"
        )
        if verbose == 2:
            print(beta_results)
        elif verbose == 1:
            print(f"beta={beta_results.root}")
        fitted_beta = beta_results.root
        return fitted_beta

    def H_func(
        self,
        beta: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
    ) -> float:
        """Returns the predicted total function

        Parameters
        ----------
        beta
            beta paramter, log of rescaling parameter
        rate_pattern
            rate pattern
        bucket_populations
            Populations in each bucket

        Returns
        -------
        float
            Predicted total given beta, bucket populations, and rate pattern
            This is matched to observed_total with rootfinding to fit beta

        """
        return np.sum(
            self.predict_count(beta, rate_pattern, bucket_populations)
        )

    def H_diff_beta(
        self,
        beta: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
    ) -> float:
        """Returns the derivative of the predicted total function with respect
        to beta

        Parameters
        ----------
        beta : float
            beta paramter, log of rescaling parameter
        rate_pattern : NDArray
            rate pattern
        bucket_populations :NDArray
            Populations in each bucket

        Returns
        -------
        float
            derivative of H_func with respect to beta
        """
        return np.sum(
            self.count_diff_beta(beta, rate_pattern, bucket_populations)
        )

    def beta_standard_error(
        self,
        fitted_beta: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
        observed_total_se: float,
    ) -> float:
        """Computes delta-method standard error estimate for beta

        Parameters
        ----------
        fitted_beta : float
            fitted value of beta. This being fitted is why we don't need what
            the original observation actually was
        rate_pattern : NDArray
            rate pattern
        bucket_populations : NDArray
            populations in each bucket
        observed_total_se : float
            standard error of the total observed quantity

        Returns
        -------
        float
            standard error of beta
        """
        error_inflation = 1 / self.H_diff_beta(
            fitted_beta, rate_pattern, bucket_populations
        )
        beta_standard_error = observed_total_se * error_inflation
        return beta_standard_error

    def split_to_rates(
        self,
        observed_total: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
        observed_total_se: float | None = None,
        lower_guess: float = -50.0,
        upper_guess: float = 50.0,
        verbose: int = 0,
        reduce_output: bool = False,
    ) -> NDArray:
        """Splits the given total to rates

        Parameters
        ----------
        observed_total
            aggregated observation
        rate_pattern
            rate pattern to assume (generalized) proportionality to
        bucket_populations
            populations in each bucket
        observed_total_se
            standard error of the observed total
        lower_guess
            Lower bound for rootfinding (we use bracketing), by default -50
        upper_guess
            Upper bound for rootfinding, by default 50
        verbose
            how much to print, 1 prints the root value,
            2 prints the entire rootfinding output, by default 0
        reduce_output
            boolean for whether or not to set groups with zero population to have zero rate

        Returns
        -------
        if observed_total_se is not given, we return
            NDArray
                predicted counts
        Otherwise we return
            NDArray
                predicted counts
            NDArray
                standard error estimates

        """
        fitted_beta = self.fit_beta(
            observed_total,
            rate_pattern,
            bucket_populations,
            lower_guess,
            upper_guess,
            verbose=verbose,
        )
        rate_point_estimates = self.predict_rate(fitted_beta, rate_pattern)

        # This is some dirty type casting, if reduce output, we set all groups with population 0 to 0
        # Otherwise, we're multiplying everything True, which gets casted to 1
        output_multiplier = ((1 - 1 * reduce_output) + bucket_populations) > 0

        if observed_total_se is not None:
            standard_errors = self.rate_standard_errors(
                fitted_beta, rate_pattern, bucket_populations, observed_total_se
            )
            return (
                rate_point_estimates * output_multiplier,
                standard_errors * output_multiplier,
            )
        return rate_point_estimates * output_multiplier

    def rate_standard_errors(
        self,
        fitted_beta: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
        observed_total_se: float,
    ) -> NDArray:
        """Computes standard errors of rate splitting outputs, only pushing
        forward uncertainty from observed_total_SE

        Parameters
        ----------
        fitted_beta
            fitted value of beta. This being fitted is why we don't need what
            the original observation actually was
        rate_pattern
            rate pattern
        bucket_populations
            populations in each bucket
        observed_total_se
            standard error of the total observed quantity

        Returns
        -------
        NDArray
            standard errors of rate splitting outputs

        """
        beta_SE = self.beta_standard_error(
            fitted_beta, rate_pattern, bucket_populations, observed_total_se
        )
        rate_standard_errors = (
            self.rate_diff_beta(fitted_beta, rate_pattern) * beta_SE
        )
        return rate_standard_errors

    def count_split_standard_errors(
        self,
        fitted_beta: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
        observed_total_se: float,
    ) -> NDArray:
        """Computes standard errors of total count splitting outputs, only
        pushing forward uncertainty from observed_total_SE

        Parameters
        ----------
        fitted_beta
            fitted value of beta. This being fitted is why we don't need what
            the original observation actually was
        rate_pattern
            rate pattern
        bucket_populations
            populations in each bucket
        observed_total_se
            standard error of the total observed quantity

        Returns
        -------
        NDArray
            standard errors of count splitting outputs

        """
        rate_se = self.rate_standard_errors(
            fitted_beta,
            rate_pattern,
            bucket_populations=bucket_populations,
            observed_total_se=observed_total_se,
        )
        return rate_se * bucket_populations

    def split_to_counts(
        self,
        observed_total: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
        observed_total_se: float | None = None,
        lower_guess: float = -50.0,
        upper_guess: float = 50.0,
        verbose: int = 0,
    ) -> NDArray:
        """Splits given total to counts

        Parameters
        ----------
        observed_total : float
            aggregated observation
        rate_pattern : NDArray
            rate pattern to assume (generalized) proportionality to
        bucket_populations : NDArray
            populations in each bucket
        observed_total_se: float, by default None
            standard error of the observed total
        lower_guess : float, optional
            Lower bound for rootfinding (we use bracketing), by default -50
        upper_guess : float, optional
            Upper bound for rootfinding, by default 50
        verbose : int, optional
            how much to print, 1 prints the root value,
            2 prints the entire rootfinding output, by default 0

        Returns
        -------
        if observed_total_se is not given, we return
            NDArray
                predicted rates
        Otherwise we return
            NDArray
                predicted rates
            NDArray
                standard error estimates

        """
        fitted_beta = self.fit_beta(
            observed_total,
            rate_pattern,
            bucket_populations,
            lower_guess,
            upper_guess,
            verbose=verbose,
        )

        fitted_beta = self.fit_beta(
            observed_total,
            rate_pattern,
            bucket_populations,
            lower_guess,
            upper_guess,
        )
        count_point_estimates = self.predict_count(
            fitted_beta, rate_pattern, bucket_populations
        )
        if observed_total_se is not None:
            standard_errors = self.count_split_standard_errors(
                fitted_beta, rate_pattern, bucket_populations, observed_total_se
            )
            return count_point_estimates, standard_errors
        return count_point_estimates

    def parameter_covariance(
        self,
        fitted_beta: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
        observed_total_se: float,
        rate_pattern_cov: NDArray,
    ) -> NDArray:
        """Computes covariance matrix of beta and rate_pattern together

        Parameters
        ----------
        fitted_beta : float
            fitted value of beta. This being fitted is why we don't need what
            the original observation actually was
        rate_pattern : NDArray
            rate pattern
        bucket_populations : NDArray
            populations in each bucket
        observed_total_se : float
            standard error of the total observed quantity
        rate_pattern_cov : NDArray
            Covariance matrix of rate pattern estimates

        Returns
        -------
        NDArray
            Joint covariance matrix with beta and rate_pattern

        """
        rate_grad = self.Hinv_rate_grad(
            fitted_beta, rate_pattern, bucket_populations
        )
        beta_pattern_cov = (rate_pattern_cov @ rate_grad).reshape(-1, 1)

        beta_var = np.array(
            [
                [
                    self.beta_standard_error(
                        fitted_beta,
                        rate_pattern,
                        bucket_populations,
                        observed_total_se,
                    )
                    ** 2
                ]
            ]
        )

        full_parameter_cov = np.block(
            [
                [
                    beta_var + rate_grad.T @ rate_pattern_cov @ rate_grad,
                    beta_pattern_cov.T,
                ],
                [beta_pattern_cov, rate_pattern_cov],
            ]
        )

        return full_parameter_cov

    def parameter_fit_jacobian(
        self,
        fitted_beta: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
    ) -> NDArray:
        """Computes jacobian matrix of H_inv with respect to beta and
        rate_pattern together

        Parameters
        ----------
        fitted_beta : float
            fitted value of beta. This being fitted is why we don't need what
            the original observation actually was
        rate_pattern : NDArray
            rate pattern
        bucket_populations : NDArray
            populations in each bucket

        Returns
        -------
        NDArray
            Joint covariance matrix with beta and rate_pattern

        """

        rate_grad = self.Hinv_rate_grad(
            fitted_beta, rate_pattern, bucket_populations
        ).reshape(-1, 1)
        beta_diff = np.array(
            [
                [
                    (
                        1
                        / self.H_diff_beta(
                            fitted_beta, rate_pattern, bucket_populations
                        )
                    )
                ]
            ]
        )
        full_parameter_jac = np.block(
            [
                [beta_diff, rate_grad.T],
                [
                    np.zeros((len(rate_pattern), 1)),
                    np.identity(len(rate_pattern)),
                ],
            ]
        )

        return full_parameter_jac

    def Hinv_rate_grad(self, beta, rate_pattern, bucket_populations) -> NDArray:
        """Gradient of H inverse (inverted with respect to beta) with respect
        to rate_pattern. The inverse is handled implicitly, as beta is the
        input, not the total.

        Parameters
        ----------
        beta : float
            beta parameter, log of rescaling parameter
        rate_pattern : NDArray
            rate pattern
        bucket_populations :NDArray
            Populations in each bucket

        Returns
        -------
        NDArray
            gradient

        """
        denominator = self.H_diff_beta(beta, rate_pattern, bucket_populations)
        # Call this at the fitted beta for the inverse derivative of H at the observed_total
        # assuming that the observed_total is H(beta), the total that beta would give!

        return (
            -1
            * self.H_rate_grad(beta, rate_pattern, bucket_populations)
            / denominator
        )

    def H_rate_grad(self, beta, rate_pattern, bucket_populations) -> NDArray:
        """Gradient of H (inverted with respect to beta) with respect to
        rate_pattern

        Parameters
        ----------
        beta : float
            beta parameter, log of rescaling parameter
        rate_pattern : NDArray
            rate pattern
        bucket_populations :NDArray
            Populations in each bucket

        Returns
        -------
        NDArray
            gradient

        """
        return (bucket_populations * self.T.diff(rate_pattern)) / (
            self.T.diff(self.T.inverse(beta + self.T(rate_pattern)))
        )

    def rate_jacobian(
        self,
        beta: float,
        rate_pattern: NDArray,
    ) -> NDArray:
        """Computes the jacobian of the predicted output rates with respect to
        both beta and the pattern

        Parameters
        ----------
        beta : float
            fitted beta parameter
        rate_pattern : NDArray
            point estimate rate pattern

        Returns
        -------
        NDArray
            Jacobian, shaped d x (d+1)

        """

        diag_scaling = np.diag(
            1 / self.T.diff(self.T.inverse(beta + self.T(rate_pattern)))
        )
        dim = len(rate_pattern)
        right_portion = np.block(
            [np.ones((dim, 1)), np.diag(self.T.diff(rate_pattern))]
        )
        return diag_scaling @ right_portion

    def count_jacobian(
        self, beta: float, rate_pattern: NDArray, bucket_populations: NDArray
    ):
        """Computes the jacobian of the predicted output counts with respect to
        both beta and the pattern

        Parameters
        ----------
        beta : float
            fitted beta parameter
        rate_pattern : NDArray
            point estimate rate pattern
        bucket_populations :NDArray
            Populations in each bucket

        Returns
        -------
        NDArray
            Jacobian, shaped d x (d+1)

        """

        rate_jac = self.rate_jacobian(beta, rate_pattern)
        diag_pops = np.diag(bucket_populations)
        return diag_pops @ rate_jac

    def rate_split_covariance_uncertainty(
        self,
        fitted_beta: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
        observed_total_se: float,
        rate_pattern_cov: NDArray,
    ) -> NDArray:
        """Computes covariance matrix of predicted rate outputs. Pushes forward
        uncertainty in both observed total, and in the rate pattern

        Parameters
        ----------
        fitted_beta : float
            fitted value of beta. This being fitted is why we don't need what
            the original observation actually was
        rate_pattern : NDArray
            rate pattern
        bucket_populations : NDArray
            populations in each bucket
        observed_total_se : float
            standard error of the total observed quantity
        rate_pattern_cov : NDArray
            Covariance matrix of rate pattern estimates

        Returns
        -------
        NDArray
            Covariance matrix of errors in splitting rate

        """
        param_covariance = self.parameter_covariance(
            fitted_beta=fitted_beta,
            rate_pattern=rate_pattern,
            bucket_populations=bucket_populations,
            observed_total_se=observed_total_se,
            rate_pattern_cov=rate_pattern_cov,
        )
        out_jac = self.rate_jacobian(
            fitted_beta,
            rate_pattern,
        )
        return out_jac @ param_covariance @ out_jac.T

    def count_split_covariance_uncertainty(
        self,
        fitted_beta: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
        observed_total_se: float,
        rate_pattern_cov: NDArray,
    ) -> NDArray:
        """Computes covariance matrix of predicted splitting count outputs
        Pushes forward uncertainty in both observed total, and in the rate
        pattern

        Parameters
        ----------
        fitted_beta : float
            fitted value of beta. This being fitted is why we don't need what
            the original observation actually was
        rate_pattern : NDArray
            rate pattern
        bucket_populations : NDArray
            populations in each bucket
        observed_total_se : float
            standard error of the total observed quantity
        rate_pattern_cov : NDArray
            Covariance matrix of rate pattern estimates

        Returns
        -------
        NDArray
            Covariance matrix of errors in splitting counts

        """
        param_covariance = self.parameter_covariance(
            fitted_beta=fitted_beta,
            rate_pattern=rate_pattern,
            bucket_populations=bucket_populations,
            observed_total_se=observed_total_se,
            rate_pattern_cov=rate_pattern_cov,
        )
        out_jac = self.count_jacobian(
            fitted_beta, rate_pattern, bucket_populations
        )
        return out_jac @ param_covariance @ out_jac.T

    def rate_split_full_jac(
        self,
        observed_total: float,
        rate_pattern: NDArray,
        bucket_populations: NDArray,
        lower_guess: float = -50,
        upper_guess: float = 50,
        verbose: int = 0,
    ) -> NDArray:
        """

        Parameters
        ----------
        observed_total : float
            aggregated observation
        rate_pattern : NDArray
            rate pattern to assume (generalized) proportionality to
        bucket_populations : NDArray
            populations in each bucket
        lower_guess : float, optional
            Lower bound for rootfinding (we use bracketing), by default -50
        upper_guess : float, optional
            Upper bound for rootfinding, by default 50
        verbose : Optional[int], optional
            how much to print, 1 prints the root value,
            2 prints the entire rootfinding output, by default 0

        Returns
        -------
        NDArray
            jacobian of prediction output

        """
        beta = self.fit_beta(
            observed_total,
            rate_pattern,
            bucket_populations,
            lower_guess=lower_guess,
            upper_guess=upper_guess,
            verbose=verbose,
        )
        # denominator = self.T.diff(self.T.inverse(self.))
        # Check if this is accomplishing what i think it is - SA 4/19
        # previously not returning anything
        return self.rate_jacobian(beta, rate_pattern)



# cat_splitter.py

from typing import Any, List, Literal
import numpy as np
from pandas import DataFrame
from pydantic import BaseModel

from pydisagg.disaggregate import split_datapoint
from pydisagg.models import RateMultiplicativeModel, LogOddsModel
from pydisagg.ihme.schema import Schema
from pydisagg.ihme.validator import (
    validate_columns,
    validate_index,
    validate_noindexdiff,
    validate_nonan,
    validate_positive,
    validate_set_uniqueness,
)


class CatDataConfig(Schema):
    """
    Configuration schema for categorical data DataFrame.
    """

    index: List[str]
    cat_group: str
    val: str
    val_sd: str

    @property
    def columns(self) -> List[str]:
        return self.index + [self.cat_group, self.val, self.val_sd]

    @property
    def val_fields(self) -> List[str]:
        return [self.val, self.val_sd]


class CatPatternConfig(Schema):
    """
    Configuration schema for the pattern DataFrame.
    """

    by: List[str]
    cat: str
    draws: List[str] = []
    val: str = "mean"
    val_sd: str = "std_err"
    prefix: str = "cat_pat_"

    @property
    def index(self) -> List[str]:
        return self.by + [self.cat]

    @property
    def columns(self) -> List[str]:
        return self.index + self.val_fields + self.draws

    @property
    def val_fields(self) -> List[str]:
        return [self.val, self.val_sd]


class CatPopulationConfig(Schema):
    """
    Configuration for the population DataFrame.
    """

    index: List[str]
    val: str
    prefix: str = "cat_pop_"

    @property
    def columns(self) -> List[str]:
        return self.index + [self.val]

    @property
    def val_fields(self) -> List[str]:
        return [self.val]


class CatSplitter(BaseModel):
    """
    Class for splitting categorical data based on pattern and population data.
    """

    data: CatDataConfig
    pattern: CatPatternConfig
    population: CatPopulationConfig

    def model_post_init(self, __context: Any) -> None:
        """
        Perform extra validation after model initialization.
        """
        if not set(self.pattern.index).issubset(self.data.index):
            raise ValueError(
                "The pattern's match criteria must be a subset of the data."
            )
        if not set(self.population.index).issubset(
            self.data.index + self.pattern.index
        ):
            raise ValueError(
                "The population's match criteria must be a subset of the data and the pattern."
            )
        if self.pattern.cat not in self.population.index:
            raise ValueError(
                "The 'target' column in the population must match the 'target' column in the data."
            )

    def parse_data(self, data: DataFrame, positive_strict: bool) -> DataFrame:
        """
        Parse and validate the input data DataFrame.
        """
        name = "While parsing data"

        # Validate required columns
        validate_columns(data, self.data.columns, name)

        # Ensure that 'cat_group' column contains lists
        data[self.data.cat_group] = data[self.data.cat_group].apply(
            lambda x: x if isinstance(x, list) else [x]
        )

        # Validate that every list in 'cat_group' contains unique elements
        validate_set_uniqueness(data, self.data.cat_group, name)

        # Explode the 'cat_group' column and rename it to match the pattern's 'cat'
        data = data.explode(self.data.cat_group).rename(
            columns={self.data.cat_group: self.pattern.cat}
        )

        # Validate index after exploding
        validate_index(data, self.data.index, name)
        validate_nonan(data, name)
        validate_positive(
            data, [self.data.val_sd], name, strict=positive_strict
        )

        return data

    def _merge_with_pattern(
        self,
        data: DataFrame,
        pattern: DataFrame,
    ) -> DataFrame:
        """
        Merge data with pattern DataFrame.
        """
        data_with_pattern = data.merge(
            pattern, on=self.pattern.index, how="left"
        )

        validate_nonan(
            data_with_pattern[
                [
                    f"{self.pattern.prefix}{col}"
                    for col in self.pattern.val_fields
                ]
            ],
            "After merging with pattern, there were NaN values created. This indicates that your pattern does not cover all the data.",
        )

        return data_with_pattern

    def parse_pattern(
        self, data: DataFrame, pattern: DataFrame, model: str
    ) -> DataFrame:
        """
        Parse and merge the pattern DataFrame with data.
        """
        name = "While parsing pattern"

        try:
            val_cols = self.pattern.val_fields
            if not all(col in pattern.columns for col in val_cols):
                if not self.pattern.draws:
                    raise ValueError(
                        f"{name}: Must provide draws for pattern if pattern.val and "
                        "pattern.val_sd are not available."
                    )
                validate_columns(pattern, self.pattern.draws, name)
                pattern[self.pattern.val] = pattern[self.pattern.draws].mean(
                    axis=1
                )
                pattern[self.pattern.val_sd] = pattern[self.pattern.draws].std(
                    axis=1
                )

            validate_columns(pattern, self.pattern.columns, name)
        except KeyError as e:
            raise KeyError(
                f"{name}: Missing columns in the pattern. Details:\n{e}"
            )

        pattern_copy = pattern.copy()
        pattern_copy = pattern_copy[
            self.pattern.index + self.pattern.val_fields
        ]
        rename_map = {
            col: f"{self.pattern.prefix}{col}"
            for col in self.pattern.val_fields
        }
        pattern_copy.rename(columns=rename_map, inplace=True)

        # Merge with pattern
        data_with_pattern = self._merge_with_pattern(data, pattern_copy)

        # Validate index differences after merging
        validate_noindexdiff(
            data,
            data_with_pattern,
            self.data.index,
            name,
        )

        return data_with_pattern

    def parse_population(
        self, data: DataFrame, population: DataFrame
    ) -> DataFrame:
        name = "Parsing Population"
        validate_columns(population, self.population.columns, name)

        population = population[self.population.columns].copy()

        validate_index(population, self.population.index, name)
        validate_nonan(population, name)

        rename_map = {
            self.population.val: f"{self.population.prefix}{self.population.val}"
        }
        population.rename(columns=rename_map, inplace=True)

        data_with_population = self._merge_with_population(data, population)

        # Ensure the prefixed population column exists
        pop_col = f"{self.population.prefix}{self.population.val}"
        if pop_col not in data_with_population.columns:
            raise KeyError(
                f"Expected column '{pop_col}' not found in merged data."
            )

        validate_nonan(
            data_with_population[[pop_col]],
            "After merging with population, there were NaN values created. This indicates that your population data does not cover all the data.",
        )
        return data_with_population

    def _merge_with_population(
        self, data: DataFrame, population: DataFrame
    ) -> DataFrame:
        """
        Merge data with population DataFrame.
        """
        data_with_population = data.merge(
            population, on=self.population.index, how="left"
        )

        return data_with_population

    def _process_group(
        self, group: DataFrame, model: str, output_type: str
    ) -> DataFrame:
        """
        Process a group of data for splitting.
        """
        observed_total = group[self.data.val].iloc[0]
        observed_total_se = group[self.data.val_sd].iloc[0]

        if len(group) == 1:
            # No need to split, assign the observed values
            group["split_result"] = observed_total
            group["split_result_se"] = observed_total_se
            group["split_flag"] = 0  # Not split
        else:
            # Need to split among multiple targets
            bucket_populations = group[
                f"{self.population.prefix}{self.population.val}"
            ].values
            rate_pattern = group[
                f"{self.pattern.prefix}{self.pattern.val}"
            ].values
            pattern_sd = group[
                f"{self.pattern.prefix}{self.pattern.val_sd}"
            ].values
            pattern_covariance = np.diag(pattern_sd**2)

            if model == "rate":
                splitting_model = RateMultiplicativeModel()
            elif model == "logodds":
                splitting_model = LogOddsModel()

            # Determine whether to normalize by population for the output type
            pop_normalize = output_type == "rate"

            # Perform splitting
            split_result, split_se = split_datapoint(
                observed_total=observed_total,
                bucket_populations=bucket_populations,
                rate_pattern=rate_pattern,
                model=splitting_model,
                output_type=output_type,
                normalize_pop_for_average_type_obs=pop_normalize,
                observed_total_se=observed_total_se,
                pattern_covariance=pattern_covariance,
            )

            # Assign results back to the group
            group["split_result"] = split_result
            group["split_result_se"] = split_se
            group["split_flag"] = 1  # Split

        return group

    def split(
        self,
        data: DataFrame,
        pattern: DataFrame,
        population: DataFrame,
        model: Literal["rate", "logodds"] = "rate",
        output_type: Literal["rate", "count"] = "rate",
    ) -> DataFrame:
        """
        Split the input data based on a specified pattern and population model.
        """
        # Validate model and output_type
        if model not in ["rate", "logodds"]:
            raise ValueError(f"Invalid model: {model}")
        if output_type not in ["rate", "count"]:
            raise ValueError(f"Invalid output_type: {output_type}")

        if self.population.prefix_status == "prefixed":
            self.population.remove_prefix()
        if self.pattern.prefix_status == "prefixed":
            self.pattern.remove_prefix()

        # Parsing input data, pattern, and population
        data = self.parse_data(data, positive_strict=True)
        data = self.parse_pattern(data, pattern, model)
        data = self.parse_population(data, population)

        # Determine grouping columns
        group_cols = self.data.index[:-1]  # Exclude 'location_id' from grouping

        # Process groups using regular groupby
        final_split_df = (
            data.groupby(group_cols, group_keys=False)
            .apply(lambda group: self._process_group(group, model, output_type))
            .reset_index(drop=True)
        )

        return final_split_df



import pytest
import pandas as pd
from pydisagg.ihme.splitter import (
    CatSplitter,
    CatDataConfig,
    CatPatternConfig,
    CatPopulationConfig,
)

# Step 1: Setup Fixtures


@pytest.fixture
def cat_data_config():
    return CatDataConfig(
        index=[
            "study_id",
            "year_id",
            "location_id",
            "sub_category",
        ],  # Include 'sub_category' in index
        cat_group="sub_category",
        val="val",
        val_sd="val_sd",
    )


@pytest.fixture
def cat_pattern_config():
    return CatPatternConfig(
        by=["year_id", "location_id"],
        cat="sub_category",
        val="pattern_val",
        val_sd="pattern_val_sd",
    )


@pytest.fixture
def cat_population_config():
    return CatPopulationConfig(
        index=["year_id", "location_id", "sub_category"],
        val="population",
    )


@pytest.fixture
def valid_data():
    return pd.DataFrame(
        {
            "study_id": [1, 2, 3],
            "year_id": [2000, 2000, 2001],
            "location_id": [10, 20, 10],
            "sub_category": [
                ["A1", "A2"],  # List of sub_categories
                ["B1", "B2"],
                ["C1", "C2"],
            ],
            "val": [100, 200, 150],
            "val_sd": [10, 20, 15],
        }
    )


@pytest.fixture
def valid_pattern():
    return pd.DataFrame(
        {
            "year_id": [2000, 2000, 2000, 2000, 2001, 2001],
            "location_id": [10, 10, 20, 20, 10, 10],
            "sub_category": ["A1", "A2", "B1", "B2", "C1", "C2"],
            "pattern_val": [0.6, 0.4, 0.7, 0.3, 0.55, 0.45],
            "pattern_val_sd": [0.06, 0.04, 0.07, 0.03, 0.055, 0.045],
        }
    )


@pytest.fixture
def valid_population():
    return pd.DataFrame(
        {
            "year_id": [2000, 2000, 2000, 2000, 2001, 2001],
            "location_id": [10, 10, 20, 20, 10, 10],
            "sub_category": ["A1", "A2", "B1", "B2", "C1", "C2"],
            "population": [5000, 3000, 7000, 3000, 5500, 4500],
        }
    )


@pytest.fixture
def cat_splitter(cat_data_config, cat_pattern_config, cat_population_config):
    return CatSplitter(
        data=cat_data_config,
        pattern=cat_pattern_config,
        population=cat_population_config,
    )


# Step 2: Write Tests for parse_data


def test_parse_data_duplicated_index(cat_splitter, valid_data):
    """Test parse_data raises an error on duplicated index."""
    duplicated_data = pd.concat([valid_data, valid_data])
    with pytest.raises(ValueError, match="has duplicated index"):
        cat_splitter.parse_data(duplicated_data, positive_strict=True)


def test_parse_data_valid(cat_splitter, valid_data):
    """Test that parse_data works correctly on valid data."""
    parsed_data = cat_splitter.parse_data(valid_data, positive_strict=True)
    assert not parsed_data.empty
    assert "val" in parsed_data.columns
    assert "val_sd" in parsed_data.columns


# Step 3: Write Tests for parse_pattern


def test_parse_pattern_valid(cat_splitter, valid_data, valid_pattern):
    """Test that parse_pattern works correctly on valid data."""
    parsed_data = cat_splitter.parse_data(valid_data, positive_strict=True)
    parsed_pattern = cat_splitter.parse_pattern(
        parsed_data, valid_pattern, model="rate"
    )
    assert not parsed_pattern.empty
    # The pattern columns are renamed with prefix 'cat_pat_'
    assert (
        f"{cat_splitter.pattern.prefix}{cat_splitter.pattern.val}"
        in parsed_pattern.columns
    )
    assert (
        f"{cat_splitter.pattern.prefix}{cat_splitter.pattern.val_sd}"
        in parsed_pattern.columns
    )


# Step 4: Write Tests for parse_population


def test_parse_population_missing_columns(
    cat_splitter, valid_data, valid_pattern, valid_population
):
    """Test parse_population raises an error when population columns are missing."""
    invalid_population = valid_population.drop(columns=["population"])
    parsed_data = cat_splitter.parse_data(valid_data, positive_strict=True)
    parsed_pattern = cat_splitter.parse_pattern(
        parsed_data, valid_pattern, model="rate"
    )
    with pytest.raises(KeyError, match="has missing columns"):
        cat_splitter.parse_population(parsed_pattern, invalid_population)


def test_parse_population_with_nan(
    cat_splitter, valid_data, valid_pattern, valid_population
):
    """Test parse_population raises an error when there are NaN values."""
    invalid_population = valid_population.copy()
    invalid_population.loc[0, "population"] = None
    parsed_data = cat_splitter.parse_data(valid_data, positive_strict=True)
    parsed_pattern = cat_splitter.parse_pattern(
        parsed_data, valid_pattern, model="rate"
    )
    with pytest.raises(ValueError, match="has NaN values"):
        cat_splitter.parse_population(parsed_pattern, invalid_population)


def test_parse_population_valid(
    cat_splitter, valid_data, valid_pattern, valid_population
):
    """Test that parse_population works correctly on valid data."""
    parsed_data = cat_splitter.parse_data(valid_data, positive_strict=True)
    parsed_pattern = cat_splitter.parse_pattern(
        parsed_data, valid_pattern, model="rate"
    )
    parsed_population = cat_splitter.parse_population(
        parsed_pattern, valid_population
    )
    assert not parsed_population.empty
    # The population column is renamed with prefix 'cat_pop_'
    pop_col = f"{cat_splitter.population.prefix}{cat_splitter.population.val}"
    assert pop_col in parsed_population.columns


# Step 5: Write Tests for the split method


def test_split_valid(cat_splitter, valid_data, valid_pattern, valid_population):
    """Test that the split method works correctly on valid data."""
    result = cat_splitter.split(
        data=valid_data,
        pattern=valid_pattern,
        population=valid_population,
        model="rate",
        output_type="rate",
    )
    assert not result.empty
    assert "split_result" in result.columns
    assert "split_result_se" in result.columns


def test_split_with_invalid_output_type(
    cat_splitter, valid_data, valid_pattern, valid_population
):
    """Test that the split method raises an error with an invalid output_type."""
    with pytest.raises(ValueError, match="Invalid output_type"):
        cat_splitter.split(
            data=valid_data,
            pattern=valid_pattern,
            population=valid_population,
            model="rate",
            output_type="invalid_output",
        )


def test_split_with_missing_population(cat_splitter, valid_data, valid_pattern):
    """Test that the split method raises an error when population data is missing."""
    with pytest.raises(
        KeyError, match="Parsing Population has missing columns"
    ):
        cat_splitter.split(
            data=valid_data,
            pattern=valid_pattern,
            population=pd.DataFrame(),  # Empty population data
            model="rate",
            output_type="rate",
        )


def test_split_with_non_matching_categories(
    cat_splitter, valid_data, valid_pattern, valid_population
):
    """Test that the split method raises an error when categories don't match."""
    invalid_population = valid_population.copy()
    invalid_population["sub_category"] = ["X1", "X2", "X1", "X2", "X1", "X2"]
    with pytest.raises(
        ValueError,
        match="After merging with population, there were NaN values created",
    ):
        cat_splitter.split(
            data=valid_data,
            pattern=valid_pattern,
            population=invalid_population,
            model="rate",
            output_type="rate",
        )



from pydisagg.models.disagg_model import DisaggModel
from pydisagg.transformations import LogOdds


class LogOddsModel(DisaggModel):
    """Produces an DisaggModel assuming multiplicativity in the odds"""

    def __init__(self) -> None:
        super().__init__(transformation=LogOdds())



import warnings
from typing import Any, Literal

import numpy as np
import pandas as pd
from pandas import DataFrame
from pydantic import BaseModel

from pydisagg.disaggregate import split_datapoint
from pydisagg.ihme.schema import Schema
from pydisagg.ihme.validator import (
    validate_columns,
    validate_index,
    validate_interval,
    validate_noindexdiff,
    validate_nonan,
    validate_pat_coverage,
    validate_positive,
    validate_realnumber,
)
from pydisagg.models import LogOddsModel, RateMultiplicativeModel


class AgeDataConfig(Schema):
    index: list[str]
    age_lwr: str
    age_upr: str
    val: str
    val_sd: str
    # sample_size: str | None

    @property
    def columns(self) -> list[str]:
        base_columns = self.index + [
            self.age_lwr,
            self.age_upr,
            self.val,
            self.val_sd,
        ]
        # if self.sample_size is not None:
        #     base_columns.append(self.sample_size)
        return list(set(base_columns))


class AgePopulationConfig(Schema):
    index: list[str]
    val: str

    prefix: str = "pop_"

    @property
    def columns(self) -> list[str]:
        return self.index + [self.val]

    @property
    def val_fields(self) -> list[str]:
        return ["val"]


class AgePatternConfig(Schema):
    by: list[str]
    age_key: str
    age_lwr: str
    age_upr: str

    draws: list[str] = []
    val: str = "val"
    val_sd: str = "val_sd"
    prefix: str = "pat_"

    @property
    def index(self) -> list[str]:
        return self.by + [self.age_key]

    @property
    def columns(self) -> list[str]:
        return self.index + [
            self.age_lwr,
            self.age_upr,
            self.val,
            self.val_sd,
        ]

    @property
    def val_fields(self) -> list[str]:
        return [
            "age_lwr",
            "age_upr",
            "val",
            "val_sd",
        ]


class AgeSplitter(BaseModel):
    data: AgeDataConfig
    pattern: AgePatternConfig
    population: AgePopulationConfig

    def model_post_init(self, __context: Any) -> None:
        """Extra validation of all the index."""
        if self.data is None or self.pattern is None or self.population is None:
            raise ValueError(
                "Data, pattern, and population configurations must all be provided."
            )
        if not set(self.pattern.by).issubset(self.data.index):
            raise ValueError("pattern.by must be a subset of data.index")
        if self.pattern.age_key in self.data.index:
            raise ValueError("pattern.age_key must not be in data.index")
        if self.pattern.age_key not in self.population.index:
            raise ValueError("pattern.age_key must be in population.index")
        if not set(self.population.index).issubset(
            self.data.index + self.pattern.index
        ):
            raise ValueError(
                "population.index must be a subset of data.index + pattern.index"
            )

    def parse_data(self, data: DataFrame, positive_strict: bool) -> DataFrame:
        name = "Parsing Data"
        validate_columns(data, self.data.columns, name)

        data = data[self.data.columns].copy()
        validate_index(data, self.data.index, name)
        validate_nonan(data, name)
        validate_positive(
            data, [self.data.val_sd], name, strict=positive_strict
        )
        validate_interval(
            data, self.data.age_lwr, self.data.age_upr, self.data.index, name
        )
        return data

    def parse_pattern(
        self, data: DataFrame, pattern: DataFrame, positive_strict: bool
    ) -> DataFrame:
        name = "Parsing Pattern"

        # Check if val and val_sd are missing, and generate them if necessary
        if not all(
            col in pattern.columns
            for col in [self.pattern.val, self.pattern.val_sd]
        ):
            if not self.pattern.draws:
                raise ValueError(
                    "Must provide draws for pattern if pattern.val and "
                    "pattern.val_sd are not available."
                )

            # Generate val and val_sd from draws
            validate_columns(pattern, self.pattern.draws, name)
            validate_realnumber(pattern, self.pattern.draws, name)
            pattern[self.pattern.val] = pattern[self.pattern.draws].mean(axis=1)
            pattern[self.pattern.val_sd] = pattern[self.pattern.draws].std(
                axis=1
            )

        # Validate columns after potential generation
        validate_columns(pattern, self.pattern.columns, name)

        # Validate for NaN values
        validate_nonan(pattern, name)

        # Validate for negative values in val_sd
        validate_positive(pattern, [self.pattern.val_sd], name, strict=False)

        # Validate interval correctness
        validate_interval(
            pattern,
            self.pattern.age_lwr,
            self.pattern.age_upr,
            self.pattern.index,
            name,
        )

        pattern_copy = pattern.copy()
        rename_map = self.pattern.apply_prefix()
        pattern_copy.rename(columns=rename_map, inplace=True)

        # Merge data with pattern
        data_with_pattern = self._merge_with_pattern(data, pattern_copy)

        # Validate index differences after merging
        validate_noindexdiff(data, data_with_pattern, self.data.index, name)

        # Validate pattern coverage
        validate_pat_coverage(
            data_with_pattern,
            self.data.age_lwr,
            self.data.age_upr,
            self.pattern.age_lwr,
            self.pattern.age_upr,
            self.data.index,
            name,
        )

        return data_with_pattern

    def _merge_with_pattern(
        self, data: DataFrame, pattern: DataFrame
    ) -> DataFrame:
        # Validate required columns
        validate_columns(pattern, self.pattern.columns, "Pattern Data")

        # Subset pattern to only the necessary columns
        pattern = pattern[self.pattern.columns].copy()

        # Proceed with merge
        data_with_pattern = data.merge(pattern, on=self.pattern.by, how="left")

        # Rest of your code (query, dropna, etc.)
        data_with_pattern = data_with_pattern.query(
            f"({self.pattern.age_lwr} >= {self.data.age_lwr} and"
            f" {self.pattern.age_lwr} < {self.data.age_upr}) or"
            f"({self.pattern.age_upr} > {self.data.age_lwr} and"
            f" {self.pattern.age_upr} <= {self.data.age_upr}) or"
            f"({self.pattern.age_lwr} <= {self.data.age_lwr} and"
            f" {self.pattern.age_upr} >= {self.data.age_upr})"
        ).dropna()

        return data_with_pattern

    def parse_population(
        self, data: DataFrame, population: DataFrame
    ) -> DataFrame:
        name = "Parsing Population"
        validate_columns(population, self.population.columns, name)

        population = population[self.population.columns].copy()

        validate_index(population, self.population.index, name)
        validate_nonan(population, name)

        rename_map = self.population.apply_prefix()
        population.rename(columns=rename_map, inplace=True)

        data_with_population = self._merge_with_population(data, population)

        validate_noindexdiff(
            data,
            data_with_population,
            self.data.index + [self.pattern.age_key],
            name,
        )
        return data_with_population

    def _merge_with_population(
        self, data: DataFrame, population: DataFrame
    ) -> DataFrame:
        data_with_population = data.merge(
            population,
            on=self.population.index,
            how="left",
        ).dropna()
        return data_with_population

    def _align_pattern_and_population(self, data: DataFrame) -> DataFrame:
        data = data.sort_values(
            self.data.index + [self.pattern.age_lwr, self.pattern.age_upr],
            ignore_index=True,
        )

        data[self.pattern.val + "_aligned"] = data[self.pattern.val]
        data[self.pattern.val_sd + "_aligned"] = data[self.pattern.val_sd]
        data[self.population.val + "_aligned"] = data[self.population.val]

        index_group = data.reset_index().groupby(self.data.index)["index"]
        index_first = index_group.first().to_list()
        index_last = index_group.last().to_list()

        data.loc[index_first, self.population.val + "_aligned"] = data.loc[
            index_first
        ].eval(
            f"{self.population.val} "
            f"/ ({self.pattern.age_upr} - {self.pattern.age_lwr}) "
            f"* ({self.pattern.age_upr} - {self.data.age_lwr})"
        )

        data.loc[index_last, self.population.val + "_aligned"] = data.loc[
            index_last
        ].eval(
            f"{self.population.val} "
            f"/ ({self.pattern.age_upr} - {self.pattern.age_lwr}) "
            f"* ({self.data.age_upr} - {self.pattern.age_lwr})"
        )

        # Not used right now, but useful in checking how we handle population partitioning
        # Can be used to split sample sizes using the pseudo-proportion
        data[self.population.val + "_total"] = data.groupby(self.data.index)[
            self.population.val + "_aligned"
        ].transform(lambda x: x.sum())
        data[self.population.val + "_proportion"] = (
            data[self.population.val + "_aligned"]
            / data[self.population.val + "_total"]
        )

        return data

    def split(
        self,
        data: DataFrame,
        pattern: DataFrame,
        population: DataFrame,
        model: Literal["rate", "logodds"] = "rate",
        output_type: Literal["rate", "count"] = "rate",
        propagate_zeros=False,
    ) -> DataFrame:
        """
        Splits the data based on the given pattern and population. The split
        results are added to the data as new columns.

        Parameters
        ----------
        data : DataFrame
            The data to be split.
        pattern : DataFrame
            The pattern to be used for splitting the data.
        population : DataFrame
            The population to be used for splitting the data.
        model : str, optional
            The model to be used for splitting the data, by default "rate".
            Can be "rate" or "logodds".
        output_type : str, optional
            The type of output to be returned, by default "rate".
        propagate_zeros : Bool, optional
            Whether to propagate pre-split zeros as post split zeros. Default false

        Returns
        -------
        DataFrame
            The two main output columns are: 'age_split_result' and 'age_split_result_se'.
            There are additional intermediate columns for sanity checks and
            calculations (have a prefix of pat_ or pop_, and a suffix of _aligned).

        """
        model_mapping = {
            "rate": RateMultiplicativeModel(),
            "logodds": LogOddsModel(),
        }

        if model not in model_mapping:
            raise ValueError(
                f"Invalid model: {model}. Expected one of: {list(model_mapping.keys())}"
            )

        model_instance = model_mapping[model]

        # If not propagating zeros,then positivity has to be strict
        if self.population.prefix_status == "prefixed":
            self.population.remove_prefix()
        if self.pattern.prefix_status == "prefixed":
            self.pattern.remove_prefix()
        data = self.parse_data(data, positive_strict=not propagate_zeros)
        data = self.parse_pattern(
            data, pattern, positive_strict=not propagate_zeros
        )
        data = self.parse_population(data, population)

        data = self._align_pattern_and_population(data)

        # where split happens
        data["age_split_result"], data["age_split_result_se"] = np.nan, np.nan
        data["age_split"] = (
            0  # Indicate that the row was not split by age initially
        )

        if propagate_zeros is True:
            data_zero = data[
                (data[self.data.val] == 0)
                | (data[self.pattern.val + "_aligned"] == 0)
            ]
            data = data[data[self.data.val] > 0]
            # Manually split zero values
            data_zero["age_split_result"] = 0.0
            data_zero["age_split_result_se"] = 0.0
            data_zero["n_ages_being_split_into"] = 1

            # Warn for all zero propagation
            num_zval = (data[self.data.val] == 0).sum()
            num_zpat = (data[self.pattern.val + "_aligned"] == 0).sum()
            num_overlap = (
                (data[self.data.val] == 0)
                * (data[self.pattern.val + "_aligned"] == 0)
            ).sum()
            if num_zval > 0:
                warnings.warn(
                    f"{num_zval} zeros produced from propagating pre-split zero values"
                )
            if num_zpat > 0:
                warnings.warn(
                    f"{num_zpat} zeros produced from propagating pattern zero values"
                )
            if num_overlap > 0:
                warnings.warn(
                    f"{num_overlap} zeros produced from this were overlapping"
                )

        data_group = data.groupby(self.data.index)
        if output_type == "count":
            pop_normalize = False
        elif output_type == "rate":
            pop_normalize = True

        n_ages_being_split_into = []

        for key, data_sub in data_group:
            split_result, SE = split_datapoint(
                observed_total=data_sub[self.data.val].iloc[0],
                bucket_populations=data_sub[
                    self.population.val + "_aligned"
                ].to_numpy(),
                rate_pattern=data_sub[self.pattern.val + "_aligned"].to_numpy(),
                model=model_instance,
                output_type=output_type,  # type: ignore, this is handeled by model_mapping
                normalize_pop_for_average_type_obs=pop_normalize,
                observed_total_se=data_sub[self.data.val_sd].iloc[0],
                pattern_covariance=np.diag(
                    data_sub[self.pattern.val_sd + "_aligned"].to_numpy() ** 2
                ),
            )
            index = data_group.groups[key]
            data.loc[index, "age_split_result"] = split_result
            data.loc[index, "age_split_result_se"] = SE
            # Check if the split values are different from the original values
            if not np.allclose(split_result, data_sub[self.data.val].iloc[0]):
                data.loc[index, "age_split"] = (
                    1  # Indicate that the row was split by age
                )
            n_ages_being_split_into.extend([len(index)] * len(index))

        if propagate_zeros is True:
            data = pd.concat([data, data_zero])
            n_ages_being_split_into.extend([1] * len(data_zero))

        self.pattern.remove_prefix()
        self.population.remove_prefix()

        data["n_ages_being_split_into"] = n_ages_being_split_into

        return data



import numpy as np
import pandas as pd
from pandas import DataFrame


def validate_columns(df: DataFrame, columns: list[str], name: str) -> None:
    """
    Validates that all specified columns are present in the DataFrame.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame to validate.
    columns : list of str
        A list of expected column names that should be present in the DataFrame.
    name : str
        A name for the DataFrame, used in error messages.

    Raises
    ------
    KeyError
        If any of the specified columns are missing from the DataFrame.
    """
    missing = [col for col in columns if col not in df.columns]
    if missing:
        error_message = (
            f"{name} has missing columns: {len(missing)} columns are missing.\n"
        )
        error_message += f"Missing columns: {', '.join(missing)}\n"
        if len(missing) > 5:
            error_message += "First 5 missing columns: \n"
        error_message += ", \n".join(missing[:5])
        error_message += "\n"
        raise KeyError(error_message)


def validate_index(df: DataFrame, index: list[str], name: str) -> None:
    """
    Validates that the DataFrame does not contain duplicate indices based on specified columns.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame to validate.
    index : list of str
        A list of column names to be used as the index for validation.
    name : str
        A name for the DataFrame, used in error messages.

    Raises
    ------
    ValueError
        If duplicate indices are found in the DataFrame based on the specified columns.
    """
    duplicated_index = pd.MultiIndex.from_frame(
        df[df[index].duplicated()][index]
    ).to_list()
    if duplicated_index:
        error_message = f"{name} has duplicated index with {len(duplicated_index)} indices \n"
        error_message += f"Index columns: ({', '.join(index)})\n"
        if len(duplicated_index) > 5:
            error_message += "First 5: \n"
        error_message += ", \n".join(str(idx) for idx in duplicated_index[:5])
        error_message += "\n"
        raise ValueError(error_message)


def validate_nonan(df: DataFrame, name: str) -> None:
    """
    Validates that the DataFrame does not contain any NaN values.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame to validate.
    name : str
        A name for the DataFrame, used in error messages.

    Raises
    ------
    ValueError
        If any NaN values are found in the DataFrame.
    """
    nan_columns = df.columns[df.isna().any(axis=0)].to_list()
    if nan_columns:
        error_message = (
            f"{name} has NaN values in {len(nan_columns)} columns. \n"
        )
        error_message += f"Columns with NaN values: {', '.join(nan_columns)}\n"
        if len(nan_columns) > 5:
            error_message += "First 5 columns with NaN values: \n"
        error_message += ", \n".join(nan_columns[:5])
        error_message += "\n"
        raise ValueError(error_message)


def validate_positive(
    df: DataFrame, columns: list[str], name: str, strict: bool = False
) -> None:
    """
    Validates that specified columns contain non-negative or strictly positive values.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame to validate.
    columns : list of str
        A list of column names to check for positive values.
    name : str
        A name for the DataFrame, used in error messages.
    strict : bool, optional
        If True, checks that values are strictly greater than zero.
        If False, checks that values are greater than or equal to zero.
        Default is False.

    Raises
    ------
    ValueError
        If any of the specified columns contain invalid (negative or zero) values.
    """
    op = "<=" if strict else "<"
    negative = [col for col in columns if df.eval(f"{col} {op} 0").any()]
    if negative:
        message = "0 or negative values in" if strict else "negative values in"
        raise ValueError(f"{name} has {message}: {negative}")


def validate_interval(
    df: DataFrame, lwr: str, upr: str, index: list[str], name: str
) -> None:
    """
    Validates that lower interval bounds are strictly less than upper bounds.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame containing interval data to validate.
    lwr : str
        The name of the column representing the lower bound of the interval.
    upr : str
        The name of the column representing the upper bound of the interval.
    index : list of str
        A list of column names to be used as the index for identifying intervals.
    name : str
        A name for the DataFrame, used in error messages.

    Raises
    ------
    ValueError
        If any lower bound is not strictly less than its corresponding upper bound.
    """
    invalid_index = pd.MultiIndex.from_frame(
        df.query(f"{lwr} >= {upr}")[index]
    ).to_list()
    if invalid_index:
        error_message = f"{name} has invalid interval with {len(invalid_index)} indices. \nLower bound must be strictly less than upper bound.\n"
        error_message += f"Index columns: ({', '.join(index)})\n"
        if len(invalid_index) > 5:
            error_message += "First 5 indices with invalid interval: \n"
        error_message += ", \n".join(str(idx) for idx in invalid_index[:5])
        error_message += "\n"
        raise ValueError(error_message)


def validate_noindexdiff(
    df_ref: DataFrame, df: DataFrame, index: list[str], name: str
) -> None:
    """
    Validates that the indices of two DataFrames match.

    Parameters
    ----------
    df_ref : pandas.DataFrame
        The reference DataFrame containing the expected indices.
    df : pandas.DataFrame
        The DataFrame to validate against the reference.
    index : list of str
        A list of column names to be used as the index for comparison.
    name : str
        A name for the validation context, used in error messages.

    Raises
    ------
    ValueError
        If there are indices in the reference DataFrame that are missing in the DataFrame to validate.
    """
    index_ref = pd.MultiIndex.from_frame(df_ref[index])
    index_to_check = pd.MultiIndex.from_frame(df[index])
    missing_index = index_ref.difference(index_to_check).to_list()

    if missing_index:
        error_message = (
            f"Missing {name} info for {len(missing_index)} indices \n"
        )
        error_message += f"Index columns: ({', '.join(index_ref.names)})\n"
        if len(missing_index) > 5:
            error_message += "First 5 missing indices: \n"
        error_message += ", \n".join(str(idx) for idx in missing_index[:5])
        error_message += "\n"
        raise ValueError(error_message)


def validate_pat_coverage(
    df: DataFrame,
    lwr: str,
    upr: str,
    pat_lwr: str,
    pat_upr: str,
    index: list[str],
    name: str,
) -> None:
    """
    Validates that the pattern intervals cover the data intervals completely without gaps or overlaps.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame containing both data intervals and pattern intervals.
    lwr : str
        The name of the column representing the data's lower bound.
    upr : str
        The name of the column representing the data's upper bound.
    pat_lwr : str
        The name of the column representing the pattern's lower bound.
    pat_upr : str
        The name of the column representing the pattern's upper bound.
    index : list of str
        A list of column names to group by when validating intervals.
    name : str
        A name for the DataFrame or validation context, used in error messages.

    Raises
    ------
    ValueError
        If the pattern intervals have gaps or overlaps, or if they do not fully cover the data intervals.
    """
    # Sort dataframe
    df = df.sort_values(index + [lwr, upr, pat_lwr, pat_upr], ignore_index=True)
    df_group = df.groupby(index)

    # Check overlap or gap in pattern
    shifted_pat_upr = df_group[pat_upr].shift(1)
    connect_index = shifted_pat_upr.notnull()
    connected = np.allclose(
        shifted_pat_upr[connect_index], df.loc[connect_index, pat_lwr]
    )

    if not connected:
        raise ValueError(
            f"{name} pattern has overlap or gap between the lower and upper "
            "bounds across categories."
        )

    # Check coverage of head and tail
    head_covered = df_group.first().eval(f"{lwr} >= {pat_lwr}").all()
    tail_covered = df_group.last().eval(f"{upr} <= {pat_upr}").all()

    if not (head_covered and tail_covered):
        raise ValueError(
            f"{name} pattern does not cover the data lower and/or upper bound"
        )


def validate_realnumber(df: DataFrame, columns: list[str], name: str) -> None:
    """
    Validates that specified columns contain real numbers, are non-zero, and are not NaN or Inf.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame containing the data to validate.
    columns : list of str
        A list of column names to validate within the DataFrame.
    name : str
        A name for the DataFrame, used in error messages.

    Raises
    ------
    ValueError
        If any column contains values that are not real numbers, are zero, or are NaN/Inf.
    """
    # Check for non-real, zero, NaN, or Inf values in the specified columns
    invalid = []
    for col in columns:
        if (
            not df[col]
            .apply(
                lambda x: isinstance(x, (int, float))
                and x != 0
                and pd.notna(x)
                and np.isfinite(x)
            )
            .all()
        ):
            invalid.append(col)

    if invalid:
        raise ValueError(f"{name} has non-real or zero values in: {invalid}")


def validate_set_uniqueness(df: DataFrame, column: str, name: str) -> None:
    """
    Validates that each list in the specified column contains unique elements.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame containing the data to validate.
    column : str
        The name of the column containing lists to validate.
    name : str
        A name for the DataFrame or validation context, used in error messages.

    Raises
    ------
    ValueError
        If any list in the specified column contains duplicate elements.
    """
    invalid_rows = df[df[column].apply(lambda x: len(x) != len(set(x)))]
    if not invalid_rows.empty:
        error_message = f"{name} has rows in column '{column}' where list elements are not unique.\n"
        error_message += (
            f"Indices of problematic rows: {invalid_rows.index.tolist()}\n"
        )
        raise ValueError(error_message)



import numpy as np
import pytest
from numpy.testing import assert_approx_equal

from pydisagg import models
from pydisagg.models.disagg_model import DisaggModel

# Need to test
# from pydisagg.disaggregate import split_datapoint

model_list = [
    models.RateMultiplicativeModel(),
    models.LogOddsModel(),
    models.LMOModel(5),
]


@pytest.mark.parametrize("model", model_list)
def test_model_consistency(model: DisaggModel):
    populations = np.array([2, 5])
    measured_total = 4.8
    measurement_SE = 1
    rate_pattern = np.array([0.2, 0.4])

    split_result = model.split_to_counts(
        measured_total,
        rate_pattern,
        populations,
    )
    beta = model.fit_beta(measured_total, rate_pattern, populations)
    split_SE_vals = model.count_split_standard_errors(
        beta,
        rate_pattern,
        populations,
        measurement_SE,
    )
    assert_approx_equal(measured_total, np.sum(split_result))
    assert_approx_equal(measurement_SE, np.sum(split_SE_vals))



from .age_splitter import (
    AgeSplitter,
    AgeDataConfig,
    AgePatternConfig,
    AgePopulationConfig,
)
from .sex_splitter import (
    SexSplitter,
    SexDataConfig,
    SexPatternConfig,
    SexPopulationConfig,
)
from .cat_splitter import (
    CatSplitter,
    CatDataConfig,
    CatPatternConfig,
    CatPopulationConfig,
)

__all__ = [
    "AgeSplitter",
    "AgeDataConfig",
    "AgePatternConfig",
    "AgePopulationConfig",
    "SexSplitter",
    "SexDataConfig",
    "SexPatternConfig",
    "SexPopulationConfig",
    "CatSplitter",
    "CatDataConfig",
    "CatPatternConfig",
    "CatPopulationConfig",
]



import pytest
from pydisagg.ihme.schema import Schema, PrefixStatus


class TestSchema(Schema):
    val: str = "value"
    val_sd: str = "value_sd"

    @property
    def val_fields(self) -> list[str]:
        return ["val", "val_sd"]


def test_apply_prefix():
    schema = TestSchema(prefix="test_")
    rename_map = schema.apply_prefix()
    assert schema.val == "test_value"
    assert schema.val_sd == "test_value_sd"
    assert schema.prefix_status == PrefixStatus.PREFIXED
    assert rename_map == {"value": "test_value", "value_sd": "test_value_sd"}


def test_apply_prefix_already_applied():
    schema = TestSchema(prefix="test_", prefix_status=PrefixStatus.PREFIXED)
    with pytest.raises(ValueError, match="prefix already applied"):
        schema.apply_prefix()


def test_remove_prefix():
    schema = TestSchema(
        prefix="test_",
        val="test_value",
        val_sd="test_value_sd",
        prefix_status=PrefixStatus.PREFIXED,
    )
    schema.remove_prefix()
    assert schema.val == "value"
    assert schema.val_sd == "value_sd"
    assert schema.prefix_status == PrefixStatus.NOT_PREFIXED


def test_remove_prefix_not_applied():
    schema = TestSchema(prefix="test_")
    schema.remove_prefix()
    assert schema.val == "value"
    assert schema.val_sd == "value_sd"
    assert schema.prefix_status == PrefixStatus.NOT_PREFIXED


def test_columns_property():
    schema = TestSchema()
    assert schema.columns == []


def test_val_fields_property():
    schema = TestSchema()
    assert schema.val_fields == ["val", "val_sd"]



from pydisagg.models.disagg_model import DisaggModel
from pydisagg.models.lmo_model import LMOModel
from pydisagg.models.log_odds_model import LogOddsModel
from pydisagg.models.rate_multiplicative_model import RateMultiplicativeModel

__all__ = ["DisaggModel", "LMOModel", "LogOddsModel", "RateMultiplicativeModel"]



import pytest
import pandas as pd
import numpy as np
from pydisagg.ihme.splitter import (
    AgeSplitter,
    AgeDataConfig,
    AgePatternConfig,
    AgePopulationConfig,
)
from pydisagg.ihme.validator import (
    validate_positive,
    validate_noindexdiff,
)


@pytest.fixture
def data():
    np.random.seed(123)
    return pd.DataFrame(
        dict(
            uid=range(10),
            sex_id=[1] * 5 + [2] * 5,
            location_id=[1, 2] * 5,
            year_id=[2010] * 10,
            age_start=[0, 5, 10, 17, 20] * 2,
            age_end=[12, 10, 22, 21, 25] * 2,
            val=5.0,
            val_sd=1.0,
        )
    )


@pytest.fixture
def pattern():
    np.random.seed(123)
    pattern_df1 = pd.DataFrame(
        dict(
            sex_id=[1] * 5 + [2] * 5,
            age_start=[0, 5, 10, 15, 20] * 2,
            age_end=[5, 10, 15, 20, 25] * 2,
            age_group_id=list(range(5)) * 2,
            draw_0=np.random.rand(10),
            draw_1=np.random.rand(10),
            draw_2=np.random.rand(10),
            year_id=[2010] * 10,
            location_id=[1] * 10,
        )
    )
    pattern_df2 = pattern_df1.copy()
    pattern_df2["location_id"] = 2
    return pd.concat([pattern_df1, pattern_df2]).reset_index(drop=True)


@pytest.fixture
def population():
    np.random.seed(123)
    sex_id = pd.DataFrame(dict(sex_id=[1, 2]))
    year_id = pd.DataFrame(dict(year_id=[2010]))
    location_id = pd.DataFrame(dict(location_id=[1, 2]))
    age_group_id = pd.DataFrame(dict(age_group_id=range(5)))

    population = (
        sex_id.merge(location_id, how="cross")
        .merge(age_group_id, how="cross")
        .merge(year_id, how="cross")
    )
    population["population"] = 1000
    return population


@pytest.fixture
def splitter(data, pattern, population):
    pattern_config = AgePatternConfig(
        by=["sex_id", "location_id", "year_id"],
        age_key="age_group_id",
        age_lwr="age_start",
        age_upr="age_end",
        draws=["draw_0", "draw_1", "draw_2"],
    )
    data_config = AgeDataConfig(
        index=["uid", "sex_id", "location_id", "year_id"],
        age_lwr="age_start",
        age_upr="age_end",
        val="val",
        val_sd="val_sd",
    )
    population_config = AgePopulationConfig(
        index=["sex_id", "location_id", "age_group_id", "year_id"],
        val="population",
    )
    return AgeSplitter(
        data=data_config, pattern=pattern_config, population=population_config
    )


def test_parse_data_success(splitter, data):
    # Should pass without exceptions
    parsed_data = splitter.parse_data(data, positive_strict=True)
    assert isinstance(parsed_data, pd.DataFrame)
    assert not parsed_data.empty


def test_parse_pattern_success(splitter, data, pattern):
    # Ensure the pattern parsing works correctly with valid input
    parsed_pattern = splitter.parse_pattern(data, pattern, positive_strict=True)
    assert isinstance(parsed_pattern, pd.DataFrame)
    assert not parsed_pattern.empty


def test_parse_data_invalid_intervals(splitter, data):
    # Introduce invalid interval (age_start >= age_end)
    data.loc[0, "age_start"] = 20
    data.loc[0, "age_end"] = 10
    with pytest.raises(ValueError):
        splitter.parse_data(data, positive_strict=True)


def test_parse_pattern_pat_coverage(splitter, data, pattern):
    # Modify pattern intervals so they don't cover the data intervals
    pattern.loc[0, "age_start"] = 50
    pattern.loc[0, "age_end"] = 60
    with pytest.raises(ValueError):
        splitter.parse_pattern(data, pattern, positive_strict=True)


def test_validate_positive():
    df = pd.DataFrame({"val_sd": [-1.0, -0.5, 0.0, 1.0, 2.0]})

    with pytest.raises(ValueError):
        validate_positive(df, ["val_sd"], "Test Pattern", strict=True)


def test_validate_noindexdiff():
    # Create two DataFrames with different indices
    df_ref = pd.DataFrame(
        {
            "uid": [0, 1, 2],
            "sex_id": [1, 1, 2],
            "location_id": [1, 1, 2],
            "year_id": [2010, 2010, 2011],
            "age_start": [0, 5, 10],
        }
    )

    df = pd.DataFrame(
        {
            "uid": [0, 1],
            "sex_id": [1, 1],
            "location_id": [1, 1],
            "year_id": [2010, 2010],
            "age_start": [0, 5],
        }
    )

    with pytest.raises(ValueError):
        validate_noindexdiff(
            df_ref, df, ["uid", "sex_id", "location_id", "year_id"], "Test Data"
        )


def test_parse_data_missing_columns(splitter, data):
    # Remove a required column
    data = data.drop(columns=["val"])
    with pytest.raises(KeyError):
        splitter.parse_data(data, positive_strict=True)


def test_parse_pattern_missing_columns(splitter, pattern):
    # Remove a required column
    pattern = pattern.drop(columns=["draw_0"])
    with pytest.raises(KeyError):
        splitter.parse_pattern(pattern, pattern, positive_strict=True)


# def test_parse_data_invalid_data_types(splitter, data):
#     # Introduce invalid data type
#     data["val"] = "invalid"
#     with pytest.raises(ValueError):
#         splitter.parse_data(data, positive_strict=True)


def test_parse_pattern_invalid_data_types(splitter, pattern):
    # Introduce invalid data type
    pattern["draw_0"] = "invalid"
    with pytest.raises(ValueError):
        splitter.parse_pattern(pattern, pattern, positive_strict=True)
